(*
(C) Copyright Andreas Viktor Hess, DTU, 2015-2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      Messages.thy
    Author:     Andreas Viktor Hess, DTU
*)

section {* Theory: Messages *}
theory Messages
imports Main "First_Order_Terms.Term"
begin

subsection {* Term-related definitions *}
fun subterms::"('a,'b) term \<Rightarrow> ('a,'b) terms" where
  "subterms (Var x) = {Var x}"
| "subterms (Fun f X) = {Fun f X} \<union> (\<Union>x \<in> (set X). subterms x)"

definition subtermeq (infix "\<sqsubseteq>" 50) where "t' \<sqsubseteq> t \<equiv> (t' \<in> subterms t)"
definition subterm (infix "\<sqsubset>" 50) where "t' \<sqsubset> t \<equiv> (t' \<sqsubseteq> t \<and> t' \<noteq> t)"

abbreviation "subterms\<^sub>s\<^sub>e\<^sub>t M \<equiv> \<Union>(subterms ` M)"

abbreviation fv where "fv \<equiv> vars_term"
lemmas fv_simps = term.simps(17,18)

definition fv\<^sub>s\<^sub>e\<^sub>t where "fv\<^sub>s\<^sub>e\<^sub>t M \<equiv> (\<Union>m \<in> M. fv m)"
definition ground where "ground M \<equiv> fv\<^sub>s\<^sub>e\<^sub>t M = {}"

declare subtermeq_def[simp]
declare subterm_def[simp]
declare ground_def[simp]
declare fv\<^sub>s\<^sub>e\<^sub>t_def[simp]


subsection {* Lemmas concerning subterms and free variables *}
lemma in_subterms_refl[simp]: "t \<in> subterms t"
by (induct t rule: subterms.induct) auto

lemma subtermI[intro]: "x \<in> subterms y \<Longrightarrow> x \<noteq> y \<Longrightarrow> x \<sqsubset> y" by simp
lemma subtermeqI[intro]: "x \<in> subterms y \<Longrightarrow> x \<sqsubseteq> y" by simp
lemma subtermeqI'[intro]: "x = y \<Longrightarrow> x \<sqsubseteq> y" using in_subterms_refl by simp
lemma subtermeqI''[intro]: "x \<in> set X \<Longrightarrow> x \<sqsubseteq> Fun f X" by force
lemma subtermeqD[dest]: "x \<sqsubseteq> y \<Longrightarrow> x \<noteq> y \<Longrightarrow> x \<sqsubset> y" by simp

lemma subterms_UnionE:
  assumes "t \<in> subterms\<^sub>s\<^sub>e\<^sub>t M"
  obtains t' where "t' \<in> M" "t \<sqsubseteq> t'"
using assms by auto

lemma finite_fv[simp]: "finite (fv t)" by (induct t) simp_all
lemma finite_fv_set[intro]: "finite M \<Longrightarrow> finite (fv\<^sub>s\<^sub>e\<^sub>t M)" by auto
lemma finite_fun_symbols[simp]: "finite (funs_term t)" by (induct t) simp_all


lemma fv_set_mono: "M \<subseteq> N \<Longrightarrow> fv\<^sub>s\<^sub>e\<^sub>t M \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t N" by auto

lemma subterms_Union_mono: "M \<subseteq> N \<Longrightarrow> subterms\<^sub>s\<^sub>e\<^sub>t M \<subseteq> subterms\<^sub>s\<^sub>e\<^sub>t N" by auto

lemma ground_subset: "M \<subseteq> N \<Longrightarrow> ground N \<Longrightarrow> ground M" unfolding ground_def by auto

lemma inv_set_fset: "finite M \<Longrightarrow> set (inv set M) = M" 
unfolding inv_def by (metis (mono_tags) finite_list someI_ex)

lemma fv_map_fv_set: "\<Union>set (map fv L) = fv\<^sub>s\<^sub>e\<^sub>t (set L)" by (induct L) auto

lemma finite_subset_Union:
  fixes A::"'a set" and f::"'a \<Rightarrow> 'b set"
  assumes "finite (\<Union>a \<in> A. f a)"
  shows "\<exists>B. finite B \<and> B \<subseteq> A \<and> (\<Union>b \<in> B. f b) = (\<Union>a \<in> A. f a)"
by (metis assms eq_iff finite_subset_image finite_UnionD)

lemma inv_set_fv: "finite M \<Longrightarrow> \<Union>set (map fv (inv set M)) = fv\<^sub>s\<^sub>e\<^sub>t M"
using fv_map_fv_set[of "inv set M"] inv_set_fset by auto

lemma fv_Fun_cons: "fv (Fun f (x#X)) = fv x \<union> fv (Fun f X)" by auto

lemma ground_subterm: "fv t = {} \<Longrightarrow> t' \<sqsubseteq> t \<Longrightarrow> fv t' = {}" by (induct t) auto

lemma ground_subterm': "fv t = {} \<Longrightarrow> t' \<sqsubset> t \<Longrightarrow> fv t' = {}" by (induct t) auto

lemma ground_subterm'': "fv t = {} \<Longrightarrow> t' \<in> subterms t \<Longrightarrow> fv t' = {}" by (induct t) auto

lemma empty_fv_not_var: "fv t = {} \<Longrightarrow> t \<noteq> Var x" by auto

lemma empty_fv_exists_fun: "fv t = {} \<Longrightarrow> \<exists>f X. t = Fun f X" by (cases t) auto

lemma vars_iff_subtermeq: "x \<in> fv t \<longleftrightarrow> Var x \<sqsubseteq> t" by (induct t) auto

lemma vars_iff_subterm_or_eq: "x \<in> fv t \<longleftrightarrow> Var x \<sqsubset> t \<or> Var x = t"
by (induct t) (auto simp add: vars_iff_subtermeq)

lemma no_var_subterm: "\<not>t \<sqsubset> Var v" by auto

lemma fun_if_subterm: "t \<sqsubset> u \<Longrightarrow> \<exists>f X. u = Fun f X" by (induct u) simp_all

lemma subtermeq_vars_subset: "M \<sqsubseteq> N \<Longrightarrow> fv M \<subseteq> fv N" by (induct N) auto

lemma fv_set_subterms_eq[simp]: "fv\<^sub>s\<^sub>e\<^sub>t M = fv\<^sub>s\<^sub>e\<^sub>t (subterms\<^sub>s\<^sub>e\<^sub>t M)"
using subtermeq_vars_subset by force

lemma fv_subset: "t \<in> M \<Longrightarrow> fv t \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t M" by auto

lemma fv_subterms: "fv t = fv\<^sub>s\<^sub>e\<^sub>t (subterms t)"
by (induct t) auto

lemma fv_subset_subterms: "t \<in> subterms\<^sub>s\<^sub>e\<^sub>t M \<Longrightarrow> fv t \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t M"
using fv_subset fv_set_subterms_eq by metis

lemma subterms_finite[simp]: "finite (subterms t)" by (induction rule: subterms.induct) auto

lemma subterms_union_finite: "finite M \<Longrightarrow> finite (\<Union>t \<in> M. subterms t)"
by (induction rule: subterms.induct) auto

lemma subterms_subset: "t' \<sqsubseteq> t \<Longrightarrow> subterms t' \<subseteq> subterms t"
by (induction rule: subterms.induct) auto

lemma subset_subterms_Union[simp]: "M \<subseteq> subterms\<^sub>s\<^sub>e\<^sub>t M" by auto

lemma in_subterms_Union: "t \<in> M \<Longrightarrow> t \<in> subterms\<^sub>s\<^sub>e\<^sub>t M" using subset_subterms_Union by blast

lemma in_subterms_subset: "t \<in> subterms t' \<Longrightarrow> subterms t \<subseteq> subterms t'"
by (simp add: subterms_subset)

lemma in_subterms_subset': "M \<subseteq> subterms t \<Longrightarrow> subterms\<^sub>s\<^sub>e\<^sub>t M \<subseteq> subterms t"
by (metis SUP_least contra_subsetD in_subterms_subset)

lemma in_subterms_subset_Union: "t \<in> subterms\<^sub>s\<^sub>e\<^sub>t M \<Longrightarrow> subterms t \<subseteq> subterms\<^sub>s\<^sub>e\<^sub>t M"
using subterms_subset by auto

lemma subtermseq_if_subterms: "t' \<sqsubset> t \<Longrightarrow> t' \<sqsubseteq> t" by simp

lemma subterm_param_split: 
  assumes "t \<sqsubset> Fun f X"
  shows "\<exists>pre x suf. t \<sqsubseteq> x \<and> X = pre@x#suf"
proof -
  obtain x where "t \<sqsubseteq> x" "x \<in> set X" using assms by auto
  then obtain pre suf where "X = pre@x#suf" "x \<notin> set pre \<or> x \<notin> set suf"
    by (meson split_list_first split_list_last)
  thus ?thesis using \<open>t \<sqsubseteq> x\<close> by auto
qed

lemma ground_iff_no_vars: "ground M \<longleftrightarrow> (\<forall>v. Var v \<notin> (\<Union>m \<in> M. subterms m))"
proof
  assume "ground M"
  hence "\<forall>v. \<forall>m \<in> M. v \<notin> fv m" by simp
  hence "\<forall>v. \<forall>m \<in> M. Var v \<notin> subterms m" by (simp add: vars_iff_subtermeq)
  thus "(\<forall>v. Var v \<notin> (\<Union>m \<in> M. subterms m))" by simp
next
  assume no_vars: "\<forall>v. Var v \<notin> (\<Union>m \<in> M. subterms m)"
  moreover
  { assume "\<not>ground M"
    then obtain v and m::"('a,'b) term" where "m \<in> M" "fv m \<noteq> {}" "v \<in> fv m"
      unfolding ground_def fv\<^sub>s\<^sub>e\<^sub>t_def by blast
    hence "Var v \<in> (subterms m)" by (simp add: vars_iff_subtermeq)
    hence "\<exists>v. Var v \<in> (\<Union>t \<in> M. subterms t)" using \<open>m \<in> M\<close> by auto
    hence False using no_vars by simp
  }
  ultimately show "ground M" by blast
qed

lemma index_Fun_subterms_subset[simp]: "i < length T \<Longrightarrow> subterms (T ! i) \<subseteq> subterms (Fun f T)"
by auto

lemma index_Fun_fv_subset[simp]: "i < length T \<Longrightarrow> fv (T ! i) \<subseteq> fv (Fun f T)"
using subtermeq_vars_subset by fastforce

lemma subterms_union_ground:
  assumes "ground M"
  shows "ground (\<Union>t \<in> M. subterms t)"
proof -
  { fix t assume "t \<in> M"
    hence "fv t = {}" using ground_iff_no_vars[of M] \<open>ground M\<close> by simp
    hence "\<forall>t' \<in> subterms t. fv t' = {}" using subtermeq_vars_subset[of _ t] by simp
    hence "ground (subterms t)" by simp
  }
  thus ?thesis by simp
qed

lemma Var_subtermeq: "t \<sqsubseteq> Var v \<Longrightarrow> t = Var v" by simp

lemma subtermeq_imp_funs_term_subset: "s \<sqsubseteq> t \<Longrightarrow> funs_term s \<subseteq> funs_term t"
by (induct t arbitrary: s) auto

lemma subterms_const: "subterms (Fun f []) = {Fun f []}" by simp

lemma subtermeq_refl[iff]: "t \<sqsubseteq> t" by (induct t rule: subterms.induct) auto

lemma subtermeq_trans[trans]: "x \<sqsubseteq> y \<Longrightarrow> y \<sqsubseteq> z \<Longrightarrow> x \<sqsubseteq> z" using subterms_subset by auto

lemma subterms_trans: "x \<in> subterms y \<Longrightarrow> y \<in> subterms z \<Longrightarrow> x \<in> subterms z"
using subterms_subset by auto

lemma subtermeq_antisym:
  assumes "x \<sqsubseteq> y" "y \<sqsubseteq> x"
  shows "x = y"
using assms
proof (induction x arbitrary: y rule: subterms.induct[case_names Var Fun])
  case (Fun f X)
  { assume "y \<noteq> Fun f X"
    then obtain t::"('a,'b) term" where "t \<in> set X" "y \<sqsubseteq> t" using \<open>y \<sqsubseteq> Fun f X\<close> by auto
    hence "Fun f X \<sqsubseteq> t" using subtermeq_trans[OF \<open>Fun f X \<sqsubseteq> y\<close>] by simp
 
    have "t \<sqsubseteq> Fun f X" using \<open>t \<in> set X\<close> by fastforce
    hence "t = Fun f X" using Fun.IH[OF \<open>t \<in> set X\<close> _ \<open>Fun f X \<sqsubseteq> t\<close>] by simp

    have "t \<sqsubseteq> y" using subtermeq_trans[OF \<open>t \<sqsubseteq> Fun f X\<close> \<open>Fun f X \<sqsubseteq> y\<close>] by simp
    hence "t = y" using Fun.IH[OF \<open>t \<in> set X\<close> _ \<open>y \<sqsubseteq> t\<close>] by simp

    have "y = Fun f X" using \<open>t = Fun f X\<close> \<open>t = y\<close> by simp
    hence False using \<open>y \<noteq> Fun f X\<close> by simp
  }
  thus ?case by auto
qed simp

lemma subterms_antisym: "x \<in> subterms y \<Longrightarrow> y \<in> subterms x \<Longrightarrow> x = y"
by (auto simp add: subtermeq_antisym)

lemma subterm_trans[trans]: "x \<sqsubset> y \<Longrightarrow> y \<sqsubset> z \<Longrightarrow> x \<sqsubset> z"
unfolding subterm_def by (meson subtermeq_def subtermeq_trans subterms_antisym)

lemma subterm_irrefl: "\<not>t \<sqsubset> t" by simp

lemma subterm_asym: "t \<sqsubset> u \<Longrightarrow> \<not>u \<sqsubset> t"
using subterm_irrefl subterm_trans by blast

lemma subterm_subtermeq_neq: "\<lbrakk>t \<sqsubset> u; u \<sqsubseteq> v\<rbrakk> \<Longrightarrow> t \<noteq> v"
using subtermeq_antisym by auto

lemma subtermeq_subterm_neq: "\<lbrakk>t \<sqsubseteq> u; u \<sqsubset> v\<rbrakk> \<Longrightarrow> t \<noteq> v"
using subtermeq_antisym by auto

lemma subtermeq_subterm_trans: "\<lbrakk>t \<sqsubseteq> u; u \<sqsubset> v\<rbrakk> \<Longrightarrow> t \<sqsubset> v"
by (metis subtermeqD subterm_trans)

lemma subterm_subtermeq_trans: "\<lbrakk>t \<sqsubset> u; u \<sqsubseteq> v\<rbrakk> \<Longrightarrow> t \<sqsubset> v"
by (metis subtermeqD subterm_trans)

lemma subterm_size_lt: "x \<sqsubset> y \<Longrightarrow> size x < size y"
using not_less_eq size_list_estimation by (induct y, simp, fastforce)

lemma in_subterms_eq: "\<lbrakk>x \<in> subterms y; y \<in> subterms x\<rbrakk> \<Longrightarrow> subterms x = subterms y"
using subterms_antisym by auto

lemma Fun_gt_params: "Fun f X \<notin> (\<Union>x \<in> set X. subterms x)"
proof -
  have "size_list size X < size (Fun f X)" by simp
  hence "Fun f X \<notin> set X" by (meson less_not_refl size_list_estimation) 
  hence "\<forall>x \<in> set X. Fun f X \<notin> subterms x \<or> x \<notin> subterms (Fun f X)"
    by (metis subterms_antisym[of "Fun f X" _])
  moreover have "\<forall>x \<in> set X. x \<in> subterms (Fun f X)" by fastforce
  ultimately show ?thesis by auto
qed

lemma params_subterms[simp]: "set X \<subseteq> subterms (Fun f X)" by auto

lemma params_subterms_Union[simp]: "subterms\<^sub>s\<^sub>e\<^sub>t (set X) \<subseteq> subterms (Fun f X)" by auto

lemma Fun_subterm_inside_params: "t \<sqsubset> Fun f X \<longleftrightarrow> t \<in> (\<Union>x \<in> (set X). subterms x)"
using Fun_gt_params by fastforce

lemma Fun_param_is_subterm: "x \<in> set X \<Longrightarrow> x \<sqsubset> Fun f X"
using Fun_subterm_inside_params by fastforce

lemma Fun_not_in_param: "x \<in> set X \<Longrightarrow> \<not>Fun f X \<sqsubset> x"
using Fun_param_is_subterm[THEN subterm_asym] .

lemma const_subterm_obtain:
  assumes "fv t = {}"
  obtains c where "Fun c [] \<sqsubseteq> t"
using assms
proof (induction t)
  case (Fun f T) thus ?case by (cases "T = []") force+
qed simp

lemma const_subterm_obtain': "fv t = {} \<Longrightarrow> \<exists>c. Fun c [] \<sqsubseteq> t"
by (metis const_subterm_obtain)

lemma subterms_singleton:
  assumes "(\<exists>v. t = Var v) \<or> (\<exists>f. t = Fun f [])"
  shows "subterms t = {t}"
using assms by (cases t) auto

lemma subtermeq_Var_const:
  assumes "s \<sqsubseteq> t"
  shows "t = Var v \<Longrightarrow> s = Var v" "t = Fun f [] \<Longrightarrow> s = Fun f []"
using assms by fastforce+

lemma subterms_singleton':
  assumes "subterms t = {t}"
  shows "(\<exists>v. t = Var v) \<or> (\<exists>f. t = Fun f [])"
proof (cases t)
  case (Fun f T)
  { fix s S assume "T = s#S"
    hence "s \<in> subterms t" using Fun by auto
    hence "s = t" using assms by auto
    hence False
      using Fun_param_is_subterm[of s "s#S" f] \<open>T = s#S\<close> Fun
      unfolding subterm_def subtermeq_def by auto
  }
  hence "T = []" by (cases T) auto
  thus ?thesis using Fun by simp
qed (simp add: assms)

lemma funs_term_subterms_eq:
  "\<And>t. \<Union>(funs_term ` subterms t) = funs_term t" 
  "\<And>A. \<Union>(funs_term ` (subterms\<^sub>s\<^sub>e\<^sub>t A)) = \<Union>(funs_term ` A)"
proof -
  show "\<And>t. \<Union>(funs_term ` subterms t) = funs_term t"
    using in_subterms_refl subtermeq_imp_funs_term_subset unfolding subtermeq_def by blast
  thus "\<And>A. \<Union>(funs_term ` (subterms\<^sub>s\<^sub>e\<^sub>t A)) = \<Union>(funs_term ` A)" by force
qed

lemmas subtermI'[intro] = Fun_param_is_subterm

lemma zip_arg_subterm:
  assumes "(s,t) \<in> set (zip X Y)"
  shows "s \<sqsubset> Fun f X" "t \<sqsubset> Fun g Y"
proof -
  from assms have "s \<in> set X" "t \<in> set Y" by (meson in_set_zipE)+
  thus "s \<sqsubset> Fun f X" "t \<sqsubset> Fun g Y" by blast+
qed

lemma zip_arg_subterm_split:
  assumes "(x,y) \<in> set (zip X Y)"
  obtains X' X'' Y' Y'' where "X = X'@x#X''" "Y = Y'@y#Y''" "length X' = length Y'"
proof -
  from assms have "\<exists>X' X'' Y' Y''. X = X'@x#X'' \<and> Y = Y'@y#Y'' \<and> length X' = length Y'"
  proof (induction Y arbitrary: X)
    case (Cons y' Y' X)
    then obtain x' X' where "X = x'#X'" by (metis empty_iff list.exhaust list.set(1) set_zip_leftD)
    show ?case
    proof (cases "(x, y) \<in> set (zip X' Y')")
      case False
      hence "x = x'" "y = y'" using Cons.prems \<open>X = x'#X'\<close> by auto
      hence "X = []@x#X' \<and> y'#Y' = []@y#Y' \<and> length [] = length []" using \<open>X = x'#X'\<close> by auto
      thus ?thesis by blast
    qed (metis \<open>X = x'#X'\<close> Cons.IH[of X'] Cons_eq_appendI list.size(4))
  qed simp
  thus ?thesis using that by blast
qed

lemma zip_arg_index:
  assumes "(x,y) \<in> set (zip X Y)"
  obtains i where "X ! i = x" "Y ! i = y" "i < length X" "i < length Y"
proof -
  obtain X1 X2 Y1 Y2 where "X = X1@x#X2" "Y = Y1@y#Y2" "length X1 = length Y1"
    using zip_arg_subterm_split[OF assms] by moura
  thus ?thesis using nth_append_length[of X1 x X2] nth_append_length[of Y1 y Y2] that by simp
qed

lemma fv_disj_Fun_subterm_param_cases:
  assumes "fv t \<inter> X = {}" "Fun f T \<in> subterms t"
  shows "T = [] \<or> (\<exists>s\<in>set T. s \<notin> Var ` X)"
proof (cases T)
  case (Cons s S)
  hence "s \<in> subterms t"
    using assms(2) subterms_trans[of _ "Fun f T" t]
    by auto
  hence "fv s \<inter> X = {}" using assms(1) fv_subterms by force
  thus ?thesis using Cons by auto
qed simp


subsection {* Other lemmas *}
lemma filter_nth: "i < length (filter P L) \<Longrightarrow> P (filter P L ! i)" using nth_mem by force 

lemma partition_index_bij:
  fixes n::nat
  obtains I k where
      "bij_betw I {0..<n} {0..<n}" "k \<le> n"
      "\<forall>i. i < k \<longrightarrow> P (I i)"
      "\<forall>i. k \<le> i \<and> i < n \<longrightarrow> \<not>(P (I i))"
proof -
  define A where "A = filter P [0..<n]"
  define B where "B = filter (\<lambda>i. \<not>P i) [0..<n]"
  define k where "k = length A"
  define I where "I = (\<lambda>n. (A@B) ! n)"

  note defs[simp] = A_def B_def k_def I_def
  
  have k1: "k \<le> n" by (metis defs(1,3) diff_le_self dual_order.trans length_filter_le length_upt)

  have "\<And>i. i < k \<Longrightarrow> P (A ! i)" by (metis defs(1,3) filter_nth)
  hence k2: "\<And>i. i < k \<Longrightarrow> P ((A@B) ! i)" by (simp add: nth_append) 

  have "\<And>i. i < length B \<Longrightarrow> \<not>(P (B ! i))" by (metis defs(2) filter_nth)
  hence "\<And>i. i < length B \<Longrightarrow> \<not>(P ((A@B) ! (k + i)))" using k_def by simp 
  hence "\<And>i. k \<le> i \<and> i < k + length B \<Longrightarrow> \<not>(P ((A@B) ! i))"
    by (metis add.commute add_less_imp_less_right le_add_diff_inverse2)
  hence k3: "\<And>i. k \<le> i \<and> i < n \<Longrightarrow> \<not>(P ((A@B) ! i))" by (simp add: sum_length_filter_compl)

  have *: "length (A@B) = n" "set (A@B) = {0..<n}" "distinct (A@B)"
    apply (metis defs(1,2) diff_zero length_append length_upt sum_length_filter_compl)
    by auto

  have I: "bij_betw I {0..<n} {0..<n}"
  proof (intro bij_betwI')
    fix x y show "x \<in> {0..<n} \<Longrightarrow> y \<in> {0..<n} \<Longrightarrow> (I x = I y) = (x = y)"
      by (metis *(1,3) defs(4) nth_eq_iff_index_eq atLeastLessThan_iff)
  next
    fix x show "x \<in> {0..<n} \<Longrightarrow> I x \<in> {0..<n}"
      by (metis *(1,2) defs(4) atLeastLessThan_iff nth_mem)
  next
    fix y show "y \<in> {0..<n} \<Longrightarrow> \<exists>x \<in> {0..<n}. y = I x"
      by (metis * defs(4) atLeast0LessThan distinct_Ex1 lessThan_iff)
  qed

  show ?thesis using k1 k2 k3 I that by simp
qed

lemma map_upt_index_eq:
  assumes "j < length T"
  shows "(map (\<lambda>i. T ! I i) [0..<length T]) ! j = T ! I j"
using assms by (simp add: map_nth)

lemma nonvar_term_has_composed_shallow_term:
  fixes t::"('f,'v) term"
  assumes "\<not>(\<exists>x. t = Var x)"
  shows "\<exists>f T. Fun f T \<sqsubseteq> t \<and> (\<forall>s \<in> set T. (\<exists>c. s = Fun c []) \<or> (\<exists>x. s = Var x))"
proof -
  let ?Q = "\<lambda>S. \<forall>s \<in> set S. (\<exists>c. s = Fun c []) \<or> (\<exists>x. s = Var x)"
  let ?P = "\<lambda>t. \<exists>g S. Fun g S \<sqsubseteq> t \<and> ?Q S"
  { fix t::"('f,'v) term"
    have "(\<exists>x. t = Var x) \<or> ?P t"
    proof (induction t)
      case (Fun h R) show ?case
      proof (cases "R = [] \<or> (\<forall>r \<in> set R. \<exists>x. r = Var x)")
        case False
        then obtain r g S where "r \<in> set R" "?P r" "Fun g S \<sqsubseteq> r" "?Q S" using Fun.IH by fast
        thus ?thesis by auto
      qed force
    qed simp
  } thus ?thesis using assms by blast
qed

end
