(*
(C) Copyright Andreas Viktor Hess, DTU, 2015-2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      Strands_And_Constraints.thy
    Author:     Andreas Viktor Hess, DTU
*)

section {* Theory: Definitions and Theorems Related to Strands and Intruder Constraints *}
theory Strands_And_Constraints
imports Messages More_Unification Intruder_Deduction
begin

subsection {* Constraints, Intruder Strands and Related Definitions *}
datatype (funs\<^sub>s\<^sub>t\<^sub>p: 'a, vars\<^sub>s\<^sub>t\<^sub>p: 'b) strand_step =
  Send       "('a,'b) term" ("send\<langle>_\<rangle>" 80)
| Receive    "('a,'b) term" ("receive\<langle>_\<rangle>" 80)
| Equality   "('a,'b) term" "('a,'b) term" ("\<langle>_ \<doteq> _\<rangle>" [80,80])
| Inequality (bvars\<^sub>s\<^sub>t\<^sub>p: "'b list") "('a,'b) term" "('a,'b) term" ("\<nexists>_\<langle>_ \<doteq> _\<rangle>" [80,80])
where
  "bvars\<^sub>s\<^sub>t\<^sub>p (Send _) = []"
| "bvars\<^sub>s\<^sub>t\<^sub>p (Receive _) = []"
| "bvars\<^sub>s\<^sub>t\<^sub>p (Equality _ _) = []"

fun trms\<^sub>s\<^sub>t\<^sub>p::"('a,'b) strand_step \<Rightarrow> ('a,'b) terms" where
  "trms\<^sub>s\<^sub>t\<^sub>p (Send t) = {t}"
| "trms\<^sub>s\<^sub>t\<^sub>p (Receive t) = {t}"
| "trms\<^sub>s\<^sub>t\<^sub>p (Equality t t') = {t,t'}"
| "trms\<^sub>s\<^sub>t\<^sub>p (Inequality _ t t') = {t,t'}"

lemma vars\<^sub>s\<^sub>t\<^sub>p_unfold[simp]: "vars\<^sub>s\<^sub>t\<^sub>p x = fv\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t\<^sub>p x) \<union> set (bvars\<^sub>s\<^sub>t\<^sub>p x)"
by (cases x) auto

definition trms\<^sub>s\<^sub>t where "trms\<^sub>s\<^sub>t S \<equiv> \<Union>(trms\<^sub>s\<^sub>t\<^sub>p ` set S)"
declare trms\<^sub>s\<^sub>t_def[simp]

type_synonym ('a,'b) strand = "('a,'b) strand_step list"
type_synonym ('a,'b) strands = "('a,'b) strand set"

(* abbreviation strand_step_prepend (infixr ":" 65) where "(b::('f,'v) strand_step):A \<equiv> b#A"
abbreviation strand_step_append (infixr ":" 65) where "A:(b::('f,'v) strand_step) \<equiv> A@[b]"
abbreviation strand_append (infixr ":" 65) where "(A::('f,'v) strand):B \<equiv> A@B" *)

definition fv\<^sub>s\<^sub>n\<^sub>d::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "fv\<^sub>s\<^sub>n\<^sub>d x \<equiv> case x of send\<langle>t\<rangle> \<Rightarrow> fv t | _ \<Rightarrow> {}"

definition fv\<^sub>r\<^sub>c\<^sub>v::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "fv\<^sub>r\<^sub>c\<^sub>v x \<equiv> case x of receive\<langle>t\<rangle> \<Rightarrow> fv t | _ \<Rightarrow> {}"

definition fv\<^sub>e\<^sub>q::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "fv\<^sub>e\<^sub>q x \<equiv> case x of (\<langle>s \<doteq> t\<rangle>) \<Rightarrow> fv s \<union> fv t | _ \<Rightarrow> {}"

definition fv_r\<^sub>e\<^sub>q::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "fv_r\<^sub>e\<^sub>q x \<equiv> case x of (\<langle>s \<doteq> t\<rangle>) \<Rightarrow> fv t | _ \<Rightarrow> {}"

definition fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x \<equiv> case x of (\<nexists>X\<langle>s \<doteq> t\<rangle>) \<Rightarrow> (fv s \<union> fv t) - set X | _ \<Rightarrow> {}"

fun fv\<^sub>s\<^sub>t\<^sub>p::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "fv\<^sub>s\<^sub>t\<^sub>p (Send t) = fv t"
| "fv\<^sub>s\<^sub>t\<^sub>p (Receive t) = fv t"
| "fv\<^sub>s\<^sub>t\<^sub>p (Equality t t') = fv t \<union> fv t'"
| "fv\<^sub>s\<^sub>t\<^sub>p (Inequality X t t') = fv t \<union> fv t' - set X"

definition fv\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> 'b set" where
  "fv\<^sub>s\<^sub>t S \<equiv> \<Union>(set (map fv\<^sub>s\<^sub>t\<^sub>p S))"

definition vars\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> 'b set" where
  "vars\<^sub>s\<^sub>t S \<equiv> \<Union>(set (map vars\<^sub>s\<^sub>t\<^sub>p S))"

definition bvars\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> 'b set" where
  "bvars\<^sub>s\<^sub>t S \<equiv> \<Union>(set (map (set \<circ> bvars\<^sub>s\<^sub>t\<^sub>p) S))"

definition nonineqvars\<^sub>s\<^sub>t\<^sub>p::"('a,'b) strand_step \<Rightarrow> 'b set" where
  "nonineqvars\<^sub>s\<^sub>t\<^sub>p x \<equiv> case x of (\<nexists>X\<langle>t \<doteq> t'\<rangle>) \<Rightarrow> {} | _ \<Rightarrow> vars\<^sub>s\<^sub>t\<^sub>p x"

definition nonineqvars\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> 'b set" where
  "nonineqvars\<^sub>s\<^sub>t S \<equiv> \<Union>set (map nonineqvars\<^sub>s\<^sub>t\<^sub>p S)"

definition wfvarsoccs\<^sub>s\<^sub>t\<^sub>p where
  "wfvarsoccs\<^sub>s\<^sub>t\<^sub>p x \<equiv> case x of send\<langle>t\<rangle> \<Rightarrow> fv t | \<langle>s \<doteq> t\<rangle> \<Rightarrow> fv s | _ \<Rightarrow> {}"

definition wfvarsoccs\<^sub>s\<^sub>t where
  "wfvarsoccs\<^sub>s\<^sub>t S \<equiv> \<Union>set (map wfvarsoccs\<^sub>s\<^sub>t\<^sub>p S)"

definition funs\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> 'a set" where
  "funs\<^sub>s\<^sub>t S \<equiv> \<Union>(set (map funs\<^sub>s\<^sub>t\<^sub>p S))"

fun eqs_rhs\<^sub>s\<^sub>t where
  "eqs_rhs\<^sub>s\<^sub>t [] = {}"
| "eqs_rhs\<^sub>s\<^sub>t (Equality t t'#S) = insert t' (eqs_rhs\<^sub>s\<^sub>t S)"
| "eqs_rhs\<^sub>s\<^sub>t (x#S) = eqs_rhs\<^sub>s\<^sub>t S"

definition subst_apply_strand_step::"('a,'b) strand_step \<Rightarrow> ('a,'b) subst \<Rightarrow> ('a,'b) strand_step"
  (infix "\<cdot>\<^sub>s\<^sub>t\<^sub>p" 51) where
  "s \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta> \<equiv> (case s of
    send\<langle>t\<rangle> \<Rightarrow> send\<langle>t \<cdot> \<theta>\<rangle>
  | receive\<langle>t\<rangle> \<Rightarrow> receive\<langle>t \<cdot> \<theta>\<rangle>
  | \<langle>t \<doteq> t'\<rangle> \<Rightarrow> \<langle>(t \<cdot> \<theta>) \<doteq> (t' \<cdot> \<theta>)\<rangle>
  | \<nexists>X\<langle>t \<doteq> t'\<rangle> \<Rightarrow> \<nexists>X\<langle>(t \<cdot> (rm_vars (set X) \<theta>)) \<doteq> (t' \<cdot> (rm_vars (set X) \<theta>))\<rangle>)"

definition subst_apply_strand::"('a,'b) strand \<Rightarrow> ('a,'b) subst \<Rightarrow> ('a,'b) strand"
  (infix "\<cdot>\<^sub>s\<^sub>t" 51) where
  "S \<cdot>\<^sub>s\<^sub>t \<theta> \<equiv> map (\<lambda>x. x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta>) S"

definition
  "ineq_model (\<I>::('a,'b) subst) X t t' \<equiv>
      (\<forall>\<delta>. subst_domain \<delta> = set X \<and> ground (subst_range \<delta>) \<longrightarrow> t \<cdot> \<delta> \<cdot> \<I> \<noteq> t' \<cdot> \<delta> \<cdot> \<I>)"

fun simple\<^sub>s\<^sub>t\<^sub>p where
  "simple\<^sub>s\<^sub>t\<^sub>p (receive\<langle>t\<rangle>) = True"
| "simple\<^sub>s\<^sub>t\<^sub>p (send\<langle>Var v\<rangle>) = True"
| "simple\<^sub>s\<^sub>t\<^sub>p (\<nexists>X\<langle>t \<doteq> t'\<rangle>) = (\<exists>\<I>. ineq_model \<I> X t t')"
| "simple\<^sub>s\<^sub>t\<^sub>p _ = False"

definition simple where "simple S \<equiv> list_all simple\<^sub>s\<^sub>t\<^sub>p S"

fun ik\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> ('a,'b) terms" where
  "ik\<^sub>s\<^sub>t [] = {}"
| "ik\<^sub>s\<^sub>t (receive\<langle>t\<rangle>#S) = insert t (ik\<^sub>s\<^sub>t S)"
| "ik\<^sub>s\<^sub>t (_#S) = ik\<^sub>s\<^sub>t S"

fun wf\<^sub>s\<^sub>t::"'b set \<Rightarrow> ('a,'b) strand \<Rightarrow> bool" where
  "wf\<^sub>s\<^sub>t V [] = True"
| "wf\<^sub>s\<^sub>t V (receive\<langle>t\<rangle>#S) = (fv t \<subseteq> V \<and> wf\<^sub>s\<^sub>t V S)"
| "wf\<^sub>s\<^sub>t V (send\<langle>t\<rangle>#S) = wf\<^sub>s\<^sub>t (V \<union> fv t) S"
| "wf\<^sub>s\<^sub>t V (\<langle>s \<doteq> t\<rangle>#S) = (fv t \<subseteq> V \<and> wf\<^sub>s\<^sub>t (V \<union> fv s) S)"
| "wf\<^sub>s\<^sub>t V (\<nexists>_\<langle>s \<doteq> t\<rangle>#S) = wf\<^sub>s\<^sub>t V S"

definition wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r::"('a,'b) strand \<Rightarrow> ('a,'b) subst \<Rightarrow> bool" where
  "wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r S \<theta> \<equiv> (wf\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> wf\<^sub>s\<^sub>t {} S \<and> subst_domain \<theta> \<inter> vars\<^sub>s\<^sub>t S = {}
                  \<and> fv_subst_range \<theta> \<inter> bvars\<^sub>s\<^sub>t S = {} \<and> fv\<^sub>s\<^sub>t S \<inter> bvars\<^sub>s\<^sub>t S = {})"

declare fv\<^sub>s\<^sub>n\<^sub>d_def[simp]
declare fv\<^sub>r\<^sub>c\<^sub>v_def[simp]
declare fv\<^sub>e\<^sub>q_def[simp]
declare fv_r\<^sub>e\<^sub>q_def[simp]
declare fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q_def[simp]
declare fv\<^sub>s\<^sub>t_def[simp]
declare vars\<^sub>s\<^sub>t_def[simp]
declare bvars\<^sub>s\<^sub>t_def[simp]
declare nonineqvars\<^sub>s\<^sub>t\<^sub>p_def[simp]
declare nonineqvars\<^sub>s\<^sub>t_def[simp]
declare wfvarsoccs\<^sub>s\<^sub>t\<^sub>p_def[simp]
declare wfvarsoccs\<^sub>s\<^sub>t_def[simp]
declare subst_apply_strand_step_def[simp]
declare subst_apply_strand_def[simp]
declare wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r_def[simp]

lemmas wf\<^sub>s\<^sub>t_induct = wf\<^sub>s\<^sub>t.induct[case_names Nil ConsRcv ConsSnd ConsEq ConsIneq]
lemmas ik\<^sub>s\<^sub>t_induct = ik\<^sub>s\<^sub>t.induct[case_names Nil ConsRcv ConsSnd ConsEq ConsIneq]
lemmas eqs_rhs\<^sub>s\<^sub>t_induct = eqs_rhs\<^sub>s\<^sub>t.induct[case_names Nil ConsEq ConsSnd ConsRcv ConsIneq]
  

subsubsection {* Lexicographical measure on strands *}
definition size\<^sub>s\<^sub>t::"('a,'b) strand \<Rightarrow> nat" where
  "size\<^sub>s\<^sub>t S \<equiv> size_list (\<lambda>x. Max (insert 0 (size ` trms\<^sub>s\<^sub>t\<^sub>p x))) S"

definition measure\<^sub>s\<^sub>t::"((('a, 'b) strand \<times> ('a,'b) subst) \<times> ('a, 'b) strand \<times> ('a,'b) subst) set"
where
  "measure\<^sub>s\<^sub>t \<equiv> measures [\<lambda>(S,\<theta>). card (fv\<^sub>s\<^sub>t S), \<lambda>(S,\<theta>). size\<^sub>s\<^sub>t S]"

declare size\<^sub>s\<^sub>t_def[simp]
declare measure\<^sub>s\<^sub>t_def[simp]

lemma measure\<^sub>s\<^sub>t_alt_def:
  "((s,x),(t,y)) \<in> measure\<^sub>s\<^sub>t =
      (card (fv\<^sub>s\<^sub>t s) < card (fv\<^sub>s\<^sub>t t) \<or> (card (fv\<^sub>s\<^sub>t s) = card (fv\<^sub>s\<^sub>t t) \<and> size\<^sub>s\<^sub>t s < size\<^sub>s\<^sub>t t))"
by simp 

lemma measure\<^sub>s\<^sub>t_trans: "trans measure\<^sub>s\<^sub>t" unfolding trans_def by auto


subsubsection {* Some lemmas *}
lemma trms\<^sub>s\<^sub>t_append: "trms\<^sub>s\<^sub>t (A@B) = trms\<^sub>s\<^sub>t A \<union> trms\<^sub>s\<^sub>t B"
by auto

lemma trm\<^sub>s\<^sub>t\<^sub>p_subst: 
  assumes "subst_domain \<theta> \<inter> set (bvars\<^sub>s\<^sub>t\<^sub>p a) = {}"
  shows "trms\<^sub>s\<^sub>t\<^sub>p a \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> = trms\<^sub>s\<^sub>t\<^sub>p (a \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta>)" (is ?A)
proof -
  have "rm_vars (set (bvars\<^sub>s\<^sub>t\<^sub>p a)) \<theta> = \<theta>" using assms by force
  thus ?A by (cases a) auto
qed

lemma trms\<^sub>s\<^sub>t_subst:
  assumes "subst_domain \<theta> \<inter> bvars\<^sub>s\<^sub>t A = {}"
  shows "trms\<^sub>s\<^sub>t A \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> = trms\<^sub>s\<^sub>t (A \<cdot>\<^sub>s\<^sub>t \<theta>)"
using assms
proof (induction A)
  case (Cons a A)
  have 1: "subst_domain \<theta> \<inter> bvars\<^sub>s\<^sub>t A = {}" "subst_domain \<theta> \<inter> set (bvars\<^sub>s\<^sub>t\<^sub>p a) = {}" using Cons.prems by auto
  hence IH: "trms\<^sub>s\<^sub>t A \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> = trms\<^sub>s\<^sub>t (A \<cdot>\<^sub>s\<^sub>t \<theta>)" using Cons.IH by simp
  
  have "trms\<^sub>s\<^sub>t (a#A) = trms\<^sub>s\<^sub>t\<^sub>p a \<union> trms\<^sub>s\<^sub>t A" by auto
  hence 2: "trms\<^sub>s\<^sub>t (a#A) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> = (trms\<^sub>s\<^sub>t\<^sub>p a \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>) \<union> (trms\<^sub>s\<^sub>t A \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>)" by (metis subst_all_union)

  have "trms\<^sub>s\<^sub>t (a#A \<cdot>\<^sub>s\<^sub>t \<theta>) = (trms\<^sub>s\<^sub>t\<^sub>p (a \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta>)) \<union> trms\<^sub>s\<^sub>t (A \<cdot>\<^sub>s\<^sub>t \<theta>)" by auto
  hence 3: "trms\<^sub>s\<^sub>t (a#A \<cdot>\<^sub>s\<^sub>t \<theta>) = (trms\<^sub>s\<^sub>t\<^sub>p a \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>) \<union> trms\<^sub>s\<^sub>t (A \<cdot>\<^sub>s\<^sub>t \<theta>)"
    using trm\<^sub>s\<^sub>t\<^sub>p_subst[OF 1(2)] by auto
  
  show ?case using IH 2 3 by metis
qed simp

lemma finite_ik\<^sub>s\<^sub>t[simp]: "finite (ik\<^sub>s\<^sub>t S)"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) simp_all

lemma finite_eqs_rhs\<^sub>s\<^sub>t[simp]: "finite (eqs_rhs\<^sub>s\<^sub>t S)"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) simp_all

lemma ik\<^sub>s\<^sub>t_is_rcv_set: "ik\<^sub>s\<^sub>t A = {t. receive\<langle>t\<rangle> \<in> set A}"
by (induct A rule: ik\<^sub>s\<^sub>t.induct) auto

lemma ik\<^sub>s\<^sub>tD[dest]: "t \<in> ik\<^sub>s\<^sub>t S \<Longrightarrow> receive\<langle>t\<rangle> \<in> set S"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma ik\<^sub>s\<^sub>tD'[dest]: "t \<in> ik\<^sub>s\<^sub>t S \<Longrightarrow> t \<in> trms\<^sub>s\<^sub>t S"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma ik\<^sub>s\<^sub>tD''[dest]: "t \<in> subterms\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<Longrightarrow> t \<in> subterms\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t S)"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma eqs_rhs\<^sub>s\<^sub>tD[dest]: "t \<in> eqs_rhs\<^sub>s\<^sub>t S \<Longrightarrow> \<exists>t'. \<langle>t' \<doteq> t\<rangle> \<in> set S"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) auto

lemma eqs_rhs\<^sub>s\<^sub>tD'[dest]: "t \<in> subterms\<^sub>s\<^sub>e\<^sub>t (eqs_rhs\<^sub>s\<^sub>t S) \<Longrightarrow> t \<in> subterms\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t S)"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) auto

lemma bvars\<^sub>s\<^sub>t_split: "bvars\<^sub>s\<^sub>t (S@S') = bvars\<^sub>s\<^sub>t S \<union> bvars\<^sub>s\<^sub>t S'"
unfolding bvars\<^sub>s\<^sub>t_def by auto

lemma bvars\<^sub>s\<^sub>t_singleton: "bvars\<^sub>s\<^sub>t [x] = set (bvars\<^sub>s\<^sub>t\<^sub>p x)"
unfolding bvars\<^sub>s\<^sub>t_def by auto

lemma strand_fv_bvars_disjointD:
  assumes "fv\<^sub>s\<^sub>t S \<inter> bvars\<^sub>s\<^sub>t S = {}" "\<nexists>X\<langle>t \<doteq> t'\<rangle> \<in> set S"
  shows "set X \<subseteq> bvars\<^sub>s\<^sub>t S" "(fv t \<union> fv t') - set X \<subseteq> fv\<^sub>s\<^sub>t S"
using assms by (induct S) fastforce+

lemma strand_fv_bvars_disjoint_unfold:
  assumes "fv\<^sub>s\<^sub>t S \<inter> bvars\<^sub>s\<^sub>t S = {}" "\<nexists>X\<langle>t \<doteq> t'\<rangle> \<in> set S" "\<nexists>Y\<langle>u \<doteq> u'\<rangle> \<in> set S"
  shows "set Y \<inter> ((fv t \<union> fv t') - set X) = {}"
proof -
  have "set X \<subseteq> bvars\<^sub>s\<^sub>t S" "set Y \<subseteq> bvars\<^sub>s\<^sub>t S"
       "(fv t \<union> fv t') - set X \<subseteq> fv\<^sub>s\<^sub>t S" "(fv u \<union> fv u') - set Y \<subseteq> fv\<^sub>s\<^sub>t S"
    using strand_fv_bvars_disjointD[OF assms(1)] assms(2,3) by auto
  thus ?thesis using assms(1) by auto
qed

lemma strand_subst_hom[iff]:
  "(S@S') \<cdot>\<^sub>s\<^sub>t \<theta> = (S \<cdot>\<^sub>s\<^sub>t \<theta>)@(S' \<cdot>\<^sub>s\<^sub>t \<theta>)" "(x#S) \<cdot>\<^sub>s\<^sub>t \<theta> = (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta>)#(S \<cdot>\<^sub>s\<^sub>t \<theta>)"
unfolding subst_apply_strand_def by auto

lemma strand_subst_comp: "fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<delta> \<circ>\<^sub>s \<theta> = ((S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t \<theta>)"
proof (induction S)
  case (Cons x S)
  have *: "fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}" "fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>) \<inter> (set (bvars\<^sub>s\<^sub>t\<^sub>p x)) = {}"
    using Cons bvars\<^sub>s\<^sub>t_split[of "[x]" S]
    apply (metis (no_types, lifting) Int_iff Un_commute append_Cons disjoint_iff_not_equal
              inf_sup_absorb self_append_conv2)
    using Cons bvars\<^sub>s\<^sub>t_split[of "[x]" S] 
    by (metis append_Cons append_self_conv2 bvars\<^sub>s\<^sub>t_singleton inf_bot_right inf_left_commute
              inf_sup_absorb) 
  hence IH: "S \<cdot>\<^sub>s\<^sub>t \<delta> \<circ>\<^sub>s \<theta> = (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t \<theta>" using Cons.IH by auto
  have "(x#S \<cdot>\<^sub>s\<^sub>t \<delta> \<circ>\<^sub>s \<theta>) = (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta> \<circ>\<^sub>s \<theta>)#(S \<cdot>\<^sub>s\<^sub>t \<delta> \<circ>\<^sub>s \<theta>)" by (metis strand_subst_hom(2))
  hence "... = (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta> \<circ>\<^sub>s \<theta>)#((S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t \<theta>)" by (metis IH)
  hence "... = ((x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta>)#((S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t \<theta>)"
    using rm_vars_comp[OF *(2)] by (cases x) simp_all
  thus ?case using IH by auto
qed simp

lemma strand_substI[intro]:
  "subst_domain \<theta> \<inter> fv\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<theta> = S"
  "subst_domain \<theta> \<inter> vars\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<theta> = S"
proof -
  show "subst_domain \<theta> \<inter> vars\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<theta> = S"
  proof (induction S)
    case (Cons x S)
    hence "S \<cdot>\<^sub>s\<^sub>t \<theta> = S" by auto
    moreover have "vars\<^sub>s\<^sub>t\<^sub>p x \<inter> subst_domain \<theta> = {}" using Cons.prems by auto
    hence "x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta> = x" using rm_vars_subst_apply_ident by (cases x) fastforce+
    ultimately show ?case by simp
  qed simp

  show "subst_domain \<theta> \<inter> fv\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<theta> = S"
  proof (induction S)
    case (Cons x S)
    hence "S \<cdot>\<^sub>s\<^sub>t \<theta> = S" by auto
    moreover have "fv\<^sub>s\<^sub>t\<^sub>p x \<inter> subst_domain \<theta> = {}" using Cons.prems by auto
    hence "x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<theta> = x" using rm_vars_subst_apply_ident by (cases x) fastforce+
    ultimately show ?case by simp
  qed simp
qed

lemma strand_substI':
  "fv\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<theta> = S"
  "vars\<^sub>s\<^sub>t S = {} \<Longrightarrow> S \<cdot>\<^sub>s\<^sub>t \<theta> = S"
by (metis inf_bot_right strand_substI(1),
    metis inf_bot_right strand_substI(2))

lemma ik_union_subset: "\<Union>(P ` ik\<^sub>s\<^sub>t S) \<subseteq> (\<Union>x \<in> (set S). \<Union>(P ` trms\<^sub>s\<^sub>t\<^sub>p x))"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma ik_snd_empty[simp]: "ik\<^sub>s\<^sub>t (map Send X) = {}"
by (induct "map Send X" arbitrary: X rule: ik\<^sub>s\<^sub>t.induct) auto

lemma ik_snd_empty'[simp]: "ik\<^sub>s\<^sub>t [send\<langle>t\<rangle>] = {}" by simp

lemma ik_append[iff]: "ik\<^sub>s\<^sub>t (S@S') = ik\<^sub>s\<^sub>t S \<union> ik\<^sub>s\<^sub>t S'" by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma ik_cons: "ik\<^sub>s\<^sub>t (x#S) = ik\<^sub>s\<^sub>t [x] \<union> ik\<^sub>s\<^sub>t S" using ik_append[of "[x]" S] by simp

lemma eqs_rhs_append[iff]: "eqs_rhs\<^sub>s\<^sub>t (S@S') = eqs_rhs\<^sub>s\<^sub>t S \<union> eqs_rhs\<^sub>s\<^sub>t S'"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) auto

lemma eqs_rcv_map_empty: "eqs_rhs\<^sub>s\<^sub>t (map Receive M) = {}"
by auto

lemma ik_rcv_map: assumes "t \<in> set L" shows "t \<in> ik\<^sub>s\<^sub>t (map Receive L)"
proof -
  { fix L L' 
    have "t \<in> ik\<^sub>s\<^sub>t [receive\<langle>t\<rangle>]" by auto
    hence "t \<in> ik\<^sub>s\<^sub>t (map Receive L@receive\<langle>t\<rangle>#map Receive L')" using ik_append by auto
    hence "t \<in> ik\<^sub>s\<^sub>t (map Receive (L@t#L'))" by auto
  }
  thus ?thesis using assms split_list_last by force 
qed

lemma ik_subst: "ik\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) = ik\<^sub>s\<^sub>t S \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>"
by (induct rule: ik\<^sub>s\<^sub>t_induct) auto

lemma ik_rcv_map': assumes "t \<in> ik\<^sub>s\<^sub>t (map Receive L)" shows "t \<in> set L"
using assms by force

lemma (in intruder_model) ik_trp_union_is_subterm_subset:
  "trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<subseteq> (\<Union>x \<in> (set S). subterms\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t\<^sub>p x))"
by (induct S rule: ik\<^sub>s\<^sub>t.induct, auto dest: trpD)

lemma (in intruder_model) trp_ik_union_append_subset[simp]:
  "trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<subseteq> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t (S@S'))"
  "trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<subseteq> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t (S'@S))"
by simp_all

lemma ik_append_subset[simp]: "ik\<^sub>s\<^sub>t S \<subseteq> ik\<^sub>s\<^sub>t (S@S')" "ik\<^sub>s\<^sub>t S' \<subseteq> ik\<^sub>s\<^sub>t (S@S')"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma eqs_rhs_append_subset[simp]: "eqs_rhs\<^sub>s\<^sub>t S \<subseteq> eqs_rhs\<^sub>s\<^sub>t (S@S')" "eqs_rhs\<^sub>s\<^sub>t S' \<subseteq> eqs_rhs\<^sub>s\<^sub>t (S@S')"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) auto

lemma (in intruder_model) trp_ik_union_append_split:
  "trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t (S@S')) = (trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)) \<union> (\<Union>(trp ` ik\<^sub>s\<^sub>t S'))"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma (in intruder_model) trp_ik_union_rcv_single:
  "trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t [receive\<langle>t\<rangle>]) = trp t"
by auto

lemma (in intruder_model) trp_ik_subtermD:
  assumes "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)"
  obtains x where "x \<in> set S" "t \<in> subterms\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t\<^sub>p x)"
using assms ik\<^sub>s\<^sub>tD trpD by auto

lemma (in intruder_model) trp_ik_subterm_exD:
  assumes "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)"
  shows "\<exists>x \<in> set S. t \<in> subterms\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t\<^sub>p x)"
using assms ik\<^sub>s\<^sub>tD trpD by force

lemma trms\<^sub>s\<^sub>t_cons: "trms\<^sub>s\<^sub>t (x#S) = trms\<^sub>s\<^sub>t\<^sub>p x \<union> trms\<^sub>s\<^sub>t S" by simp

lemma trm_strand_subst_cong:
  "t \<in> trms\<^sub>s\<^sub>t S \<Longrightarrow> t \<cdot> \<delta> \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)
    \<or> (\<exists>X t' t''. \<nexists>X\<langle>t' \<doteq> t''\<rangle> \<in> set S \<and> t \<cdot> rm_vars (set X) \<delta> \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>))"
  "t \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<Longrightarrow> (\<exists>t'. t = t' \<cdot> \<delta> \<and> t' \<in> trms\<^sub>s\<^sub>t S)
    \<or> (\<exists>X t' t''. \<nexists>X\<langle>t' \<doteq> t''\<rangle> \<in> set S \<and> (t = t' \<cdot> rm_vars (set X) \<delta> \<or> t = t'' \<cdot> rm_vars (set X) \<delta>))"
proof -
  let ?P = "\<lambda>t \<delta> S. t \<cdot> \<delta> \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)
    \<or> (\<exists>X t' t''. \<nexists>X\<langle>t' \<doteq> t''\<rangle> \<in> set S \<and> t \<cdot> rm_vars (set X) \<delta> \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>))"
  let ?Q = "\<lambda>t \<delta> S. (\<exists>t'. t = t' \<cdot> \<delta> \<and> t' \<in> trms\<^sub>s\<^sub>t S)
    \<or> (\<exists>X t' t''. \<nexists>X\<langle>t' \<doteq> t''\<rangle> \<in> set S \<and> (t = t' \<cdot> rm_vars (set X) \<delta> \<or> t = t'' \<cdot> rm_vars (set X) \<delta>))"

  show "t \<in> trms\<^sub>s\<^sub>t S \<Longrightarrow> ?P t \<delta> S"
  proof (induction S)
    case (Cons x S) show ?case
    proof (cases "t \<in> trms\<^sub>s\<^sub>t S")
      case True
      hence "?P t \<delta> S" using Cons by simp
      thus ?thesis
      apply (cases x)
        by (metis (no_types, lifting) Un_iff list.set_intros(2) strand_subst_hom(2) trms\<^sub>s\<^sub>t_cons)+
    next
      case False
      hence **: "t \<in> trms\<^sub>s\<^sub>t\<^sub>p x" using Cons.prems by auto
      thus ?thesis
      proof (cases x)
        case (Inequality X s s')
        hence "t \<cdot> rm_vars (set X) \<delta> \<in> trms\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)"
          using ** by auto
        moreover have "\<nexists>X\<langle>s \<doteq> s'\<rangle> \<in> set (x#S)" using Inequality by simp
        ultimately show ?thesis using Inequality by force
      qed auto
    qed
  qed simp

  show "t \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<Longrightarrow> ?Q t \<delta> S"
  proof (induction S)
    case (Cons x S) show ?case
    proof (cases "t \<in> trms\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)")
      case True
      hence "?Q t \<delta> S" using Cons by simp
      thus ?thesis by (cases x) auto
    next
      case False
      hence **: "t \<in> trms\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)" using Cons.prems by auto
      thus ?thesis
      proof (cases x)
        case (Inequality X s s')
        hence "t \<in> trms\<^sub>s\<^sub>t\<^sub>p x \<cdot>\<^sub>s\<^sub>e\<^sub>t rm_vars (set X) \<delta>"
          using ** by auto
        moreover have "\<nexists>X\<langle>s \<doteq> s'\<rangle> \<in> set (x#S)" using Inequality by simp
        ultimately show ?thesis using Inequality by auto
      qed auto
    qed
  qed simp
qed

lemma (in intruder_model) wf_trm_subst_rm_vars:
  "wf\<^sub>t\<^sub>r\<^sub>m (\<delta> v) \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m (rm_vars X \<delta> v)"
by auto


subsection {* Lemmas: Free Variables of Strands *}
lemma fv_trm_snd_rcv: "fv\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t\<^sub>p (send\<langle>t\<rangle>)) = fv t" "fv\<^sub>s\<^sub>e\<^sub>t (trms\<^sub>s\<^sub>t\<^sub>p (receive\<langle>t\<rangle>)) = fv t"
by simp_all

lemma in_strand_fv_subset: "x \<in> set S \<Longrightarrow> vars\<^sub>s\<^sub>t\<^sub>p x \<subseteq> vars\<^sub>s\<^sub>t S" by fastforce
lemma in_strand_fv_subset_snd: "send\<langle>t\<rangle> \<in> set S \<Longrightarrow> fv t \<subseteq> \<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)" by auto
lemma in_strand_fv_subset_rcv: "receive\<langle>t\<rangle> \<in> set S \<Longrightarrow> fv t \<subseteq> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)" by auto

lemma finite_trms\<^sub>s\<^sub>t\<^sub>p[simp]: "finite (trms\<^sub>s\<^sub>t\<^sub>p x)" by (cases x) auto
lemma finite_vars\<^sub>s\<^sub>t\<^sub>p[simp]: "finite (vars\<^sub>s\<^sub>t\<^sub>p x)" using finite_fv by auto
lemma finite_bvars\<^sub>s\<^sub>t\<^sub>p[simp]: "finite (set (bvars\<^sub>s\<^sub>t\<^sub>p x))" by rule
lemma finite_fv\<^sub>s\<^sub>n\<^sub>d[simp]: "finite (fv\<^sub>s\<^sub>n\<^sub>d x)" using finite_fv by (cases x) auto
lemma finite_fv\<^sub>r\<^sub>c\<^sub>v[simp]: "finite (fv\<^sub>r\<^sub>c\<^sub>v x)" using finite_fv by (cases x) auto
lemma finite_fv\<^sub>s\<^sub>t\<^sub>p[simp]: "finite (fv\<^sub>s\<^sub>t\<^sub>p x)" using finite_fv by (cases x) auto
lemma finite_vars\<^sub>s\<^sub>t[simp]: "finite (vars\<^sub>s\<^sub>t S)" by simp
lemma finite_bvars\<^sub>s\<^sub>t[simp]: "finite (bvars\<^sub>s\<^sub>t S)" by simp
lemma finite_fv\<^sub>s\<^sub>t[simp]: "finite (fv\<^sub>s\<^sub>t S)" by simp
lemma finite_nonineqvars\<^sub>s\<^sub>t\<^sub>p[simp]: "finite (nonineqvars\<^sub>s\<^sub>t\<^sub>p x)" by (cases x) auto
lemma finite_nonineqvars\<^sub>s\<^sub>t[simp]: "finite (nonineqvars\<^sub>s\<^sub>t S)" using finite_nonineqvars\<^sub>s\<^sub>t\<^sub>p by auto
lemma finite_wfvarsoccs\<^sub>s\<^sub>t\<^sub>p[simp]: "finite (wfvarsoccs\<^sub>s\<^sub>t\<^sub>p x)" by (cases x) auto
lemma finite_wfvarsoccs\<^sub>s\<^sub>t[simp]: "finite (wfvarsoccs\<^sub>s\<^sub>t S)" using finite_wfvarsoccs\<^sub>s\<^sub>t\<^sub>p by auto

lemma fv\<^sub>s\<^sub>n\<^sub>dE:
  assumes "v \<in> \<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)"
  obtains t where "send\<langle>t\<rangle> \<in> set S" "v \<in> fv t"
using assms
apply (induct S, simp_all)
by (metis empty_iff strand_step.case_eq_if strand_step.collapse(1))

lemma fv\<^sub>r\<^sub>c\<^sub>vE:
  assumes "v \<in> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)"
  obtains t where "receive\<langle>t\<rangle> \<in> set S" "v \<in> fv t"
using assms
apply (induct S, simp_all)
by (metis empty_iff strand_step.case_eq_if strand_step.collapse(2))

lemma vars\<^sub>s\<^sub>t\<^sub>pI[intro]: "x \<in> fv\<^sub>s\<^sub>t\<^sub>p s \<Longrightarrow> x \<in> vars\<^sub>s\<^sub>t\<^sub>p s"
by (induct s rule: fv\<^sub>s\<^sub>t\<^sub>p.induct) auto

lemma vars\<^sub>s\<^sub>tI[intro]: "x \<in> fv\<^sub>s\<^sub>t S \<Longrightarrow> x \<in> vars\<^sub>s\<^sub>t S" using vars\<^sub>s\<^sub>t\<^sub>pI by fastforce

lemma fv\<^sub>s\<^sub>t_subset_vars\<^sub>s\<^sub>t[simp]: "fv\<^sub>s\<^sub>t S \<subseteq> vars\<^sub>s\<^sub>t S" using vars\<^sub>s\<^sub>tI by force

lemma vars\<^sub>s\<^sub>t_is_fv\<^sub>s\<^sub>t_bvars\<^sub>s\<^sub>t: "vars\<^sub>s\<^sub>t S = fv\<^sub>s\<^sub>t S \<union> bvars\<^sub>s\<^sub>t S"
proof (induction S)
  case (Cons x S) thus ?case by (cases x) fastforce+
qed simp

lemma vars_st_snd_map: "vars\<^sub>s\<^sub>t (map Send X) = fv (Fun f X)" by auto
lemma vars_st_rcv_map: "vars\<^sub>s\<^sub>t (map Receive X) = fv (Fun f X)" by auto

lemma vars_snd_rcv_union: "vars\<^sub>s\<^sub>t\<^sub>p x = fv\<^sub>s\<^sub>n\<^sub>d x \<union> fv\<^sub>r\<^sub>c\<^sub>v x \<union> fv\<^sub>e\<^sub>q x \<union> fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x \<union> set (bvars\<^sub>s\<^sub>t\<^sub>p x)"
by (cases x) simp_all

lemma fv_snd_rcv_union: "fv\<^sub>s\<^sub>t\<^sub>p x = fv\<^sub>s\<^sub>n\<^sub>d x \<union> fv\<^sub>r\<^sub>c\<^sub>v x \<union> fv\<^sub>e\<^sub>q x \<union> fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x"
by (cases x) simp_all

lemma fv_snd_rcv_empty[simp]: "fv\<^sub>s\<^sub>n\<^sub>d x = {} \<or> fv\<^sub>r\<^sub>c\<^sub>v x = {}" by (cases x) simp_all

lemma vars_snd_rcv_strand[iff]:
  "vars\<^sub>s\<^sub>t (S::('a,'b) strand) =
    (\<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)) \<union> (\<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)) \<union> (\<Union>set (map fv\<^sub>e\<^sub>q S)) \<union> (\<Union>set (map fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q S))
    \<union> bvars\<^sub>s\<^sub>t S"
unfolding bvars\<^sub>s\<^sub>t_def
proof (induction S)
  case (Cons x S)
  have "\<And>s V. vars\<^sub>s\<^sub>t\<^sub>p (s::('a,'b) strand_step) \<union> V = 
                fv\<^sub>s\<^sub>n\<^sub>d s \<union> fv\<^sub>r\<^sub>c\<^sub>v s \<union> fv\<^sub>e\<^sub>q s \<union> fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q s \<union> set (bvars\<^sub>s\<^sub>t\<^sub>p s) \<union> V"
    by (metis vars_snd_rcv_union)
  thus ?case using Cons.IH by (auto simp add: sup_assoc sup_left_commute)
qed simp

lemma fv_snd_rcv_strand[iff]:
  "fv\<^sub>s\<^sub>t (S::('a,'b) strand) =
    (\<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)) \<union> (\<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)) \<union> (\<Union>set (map fv\<^sub>e\<^sub>q S)) \<union> (\<Union>set (map fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q S))"
unfolding bvars\<^sub>s\<^sub>t_def
proof (induction S)
  case (Cons x S)
  have "\<And>s V. fv\<^sub>s\<^sub>t\<^sub>p (s::('a,'b) strand_step) \<union> V = 
                fv\<^sub>s\<^sub>n\<^sub>d s \<union> fv\<^sub>r\<^sub>c\<^sub>v s \<union> fv\<^sub>e\<^sub>q s \<union> fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q s \<union> V"
    by (metis fv_snd_rcv_union)
  thus ?case using Cons.IH by (auto simp add: sup_assoc sup_left_commute)
qed simp

lemma vars_snd_rcv_strand2[iff]:
  "nonineqvars\<^sub>s\<^sub>t (S::('a,'b) strand) = (\<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)) \<union> (\<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)) \<union> (\<Union>set (map fv\<^sub>e\<^sub>q S))"
by (induction S) (auto split: strand_step.split)

lemma fv_snd_rcv_strand_subset[simp]:
  "\<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S) \<subseteq> fv\<^sub>s\<^sub>t S" "\<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S) \<subseteq> fv\<^sub>s\<^sub>t S"
  "\<Union>set (map fv\<^sub>e\<^sub>q S) \<subseteq> fv\<^sub>s\<^sub>t S" "\<Union>set (map fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q S) \<subseteq> fv\<^sub>s\<^sub>t S"
  "wfvarsoccs\<^sub>s\<^sub>t S \<subseteq> fv\<^sub>s\<^sub>t S"
proof -
  show "\<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S) \<subseteq> fv\<^sub>s\<^sub>t S" "\<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S) \<subseteq> fv\<^sub>s\<^sub>t S"
       "\<Union>set (map fv\<^sub>e\<^sub>q S) \<subseteq> fv\<^sub>s\<^sub>t S" "\<Union>set (map fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q S) \<subseteq> fv\<^sub>s\<^sub>t S"
    using fv_snd_rcv_strand[of S] by auto
  show "wfvarsoccs\<^sub>s\<^sub>t S \<subseteq> fv\<^sub>s\<^sub>t S" by (induct S) (auto split: strand_step.split)
qed

lemma vars_snd_rcv_strand_subset2[simp]:
  "\<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S) \<subseteq> nonineqvars\<^sub>s\<^sub>t S" "\<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S) \<subseteq> nonineqvars\<^sub>s\<^sub>t S"
  "\<Union>set (map fv\<^sub>e\<^sub>q S) \<subseteq> nonineqvars\<^sub>s\<^sub>t S" "wfvarsoccs\<^sub>s\<^sub>t S \<subseteq> nonineqvars\<^sub>s\<^sub>t S"
by (induction S) (auto split: strand_step.split)

lemma nonineqvars\<^sub>s\<^sub>t_subset_vars\<^sub>s\<^sub>t: "nonineqvars\<^sub>s\<^sub>t S \<subseteq> vars\<^sub>s\<^sub>t S"
by (induction S) (auto split: strand_step.split)

lemma subst_sends_strand_step_fv_to_img: "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> fv\<^sub>s\<^sub>t\<^sub>p x \<union> fv_subst_range \<delta>" 
using subst_sends_fv_to_img[of _ \<delta>]
proof (cases x)
  case (Inequality X t t')
  let ?\<theta> = "rm_vars (set X) \<delta>"
  have "fv (t \<cdot> ?\<theta>) \<subseteq> fv t \<union> fv_subst_range ?\<theta>" "fv (t' \<cdot> ?\<theta>) \<subseteq> fv t' \<union> fv_subst_range ?\<theta>"
    using Inequality subst_sends_fv_to_img[of _ ?\<theta>] by simp_all
  hence "fv (t \<cdot> ?\<theta>) \<subseteq> fv t \<union> fv_subst_range \<delta>" "fv (t' \<cdot> ?\<theta>) \<subseteq> fv t' \<union> fv_subst_range \<delta>"
    using rm_vars_img_subset[of "set X" \<delta>] fv_set_mono by blast+
  thus ?thesis using Inequality by auto
qed auto

lemma subst_sends_strand_fv_to_img: "fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<subseteq> fv\<^sub>s\<^sub>t S \<union> fv_subst_range \<delta>" 
proof (induction S)
  case (Cons x S)
  have *: "fv\<^sub>s\<^sub>t (x#S \<cdot>\<^sub>s\<^sub>t \<delta>) = fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<union> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
          "fv\<^sub>s\<^sub>t (x#S) \<union> fv_subst_range \<delta> = fv\<^sub>s\<^sub>t\<^sub>p x \<union> fv\<^sub>s\<^sub>t S \<union> fv_subst_range \<delta>"
    by auto
  thus ?case using Cons.IH subst_sends_strand_step_fv_to_img[of x \<delta>] by auto
qed simp

lemma ineq_apply_subst:
  assumes "subst_domain \<delta> \<inter> set X = {}"
  shows "(\<nexists>X\<langle>t \<doteq> t'\<rangle>) \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta> = \<nexists>X\<langle>(t \<cdot> \<delta>) \<doteq> (t' \<cdot> \<delta>)\<rangle>"
using rm_vars_apply'[OF assms] by simp

lemma fv_strand_step_subst:
  assumes "P = fv\<^sub>s\<^sub>t\<^sub>p \<or> P = fv\<^sub>r\<^sub>c\<^sub>v \<or> P = fv\<^sub>s\<^sub>n\<^sub>d \<or> P = fv\<^sub>e\<^sub>q \<or> P = fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q"
  and "set (bvars\<^sub>s\<^sub>t\<^sub>p x) \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (P x)) = P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)"
proof (cases x)
  case (Send t)
  hence "vars\<^sub>s\<^sub>t\<^sub>p x = fv t" "fv\<^sub>s\<^sub>n\<^sub>d x = fv t" by auto
  thus ?thesis using assms Send subst_apply_fv_unfold[of _ \<delta>] by auto
next
  case (Receive t)
  hence "vars\<^sub>s\<^sub>t\<^sub>p x = fv t" "fv\<^sub>r\<^sub>c\<^sub>v x = fv t" by auto
  thus ?thesis using assms Receive subst_apply_fv_unfold[of _ \<delta>] by auto
next
  case (Equality t t')
  hence "vars\<^sub>s\<^sub>t\<^sub>p x = fv t \<union> fv t'" "fv\<^sub>e\<^sub>q x = fv t \<union> fv t'" by auto
  thus ?thesis using assms Equality subst_apply_fv_unfold[of _ \<delta>] by auto
next
  case (Inequality X t t')
  hence *: "set X \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
           "x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta> = \<nexists>X\<langle>(t \<cdot> \<delta>) \<doteq> (t' \<cdot> \<delta>)\<rangle>" 
    using assms ineq_apply_subst[of \<delta> X t t'] by auto
  have "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = fv t \<union> fv t' - set X" using Inequality by auto
  hence "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x) = fv\<^sub>s\<^sub>e\<^sub>t (\<delta>`(fv t - set X)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta>`(fv t' - set X))" by auto
  hence "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x) = fv\<^sub>s\<^sub>e\<^sub>t (\<delta>`(fv t)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta>`(fv t')) - set X"
    using Inequality fv\<^sub>s\<^sub>e\<^sub>t_subst_img_eq[OF *(1)] by auto
  hence "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
    using subst_apply_fv_unfold[of _ \<delta>] Inequality by simp
  moreover have "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = (fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)) - set X" using *(2) by auto
  ultimately have "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x) = fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)"
    using assms(1) Inequality subst_apply_fv_unfold[of _ \<delta>] by simp
  thus ?thesis
    using assms(1) Inequality subst_apply_fv_unfold[of _ \<delta>] *(2) fv_snd_rcv_union
    by (smt Sup_bot_conv(1) fv\<^sub>e\<^sub>q_def fv\<^sub>r\<^sub>c\<^sub>v_def fv\<^sub>s\<^sub>n\<^sub>d_def imageE list.map(1) list.set(1) set_map
            strand_step.case(4) sup_bot.left_neutral term.set(4) vars_term_subst)
qed

lemma fv_strand_subst:
  assumes "P = fv\<^sub>s\<^sub>t\<^sub>p \<or> P = fv\<^sub>r\<^sub>c\<^sub>v \<or> P = fv\<^sub>s\<^sub>n\<^sub>d \<or> P = fv\<^sub>e\<^sub>q \<or> P = fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q"
  and "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (\<Union>set (map P S))) = \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>))"
using assms(2)
proof (induction S)
  case (Cons x S)
  hence *: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
           "set (bvars\<^sub>s\<^sub>t\<^sub>p x) \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
    unfolding bvars\<^sub>s\<^sub>t_def by force+
  hence **: "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` P x) = P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)" using fv_strand_step_subst[OF assms(1), of x \<delta>] by auto
  have "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (\<Union>set (map P (x#S)))) = fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` P x) \<union> (\<Union>set (map P ((S \<cdot>\<^sub>s\<^sub>t \<delta>))))"
    using Cons unfolding bvars\<^sub>s\<^sub>t_def by force+
  hence "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (\<Union>set (map P (x#S)))) = P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (\<Union>set (map P S)))"
    using ** by simp
  thus ?case using Cons.IH[OF *(1)] unfolding bvars\<^sub>s\<^sub>t_def by simp
qed simp

lemma fv_strand_subst2:
  assumes "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (nonineqvars\<^sub>s\<^sub>t S)) = nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
by (metis (no_types, lifting) assms fv\<^sub>s\<^sub>e\<^sub>t_def vars_snd_rcv_strand2 fv_strand_subst UN_Un image_Un)

lemma fv_strand_subst':
  assumes "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (fv\<^sub>s\<^sub>t S)) = fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
by (metis assms fv_strand_subst fv\<^sub>s\<^sub>t_def)

lemma strand_map_set_subst:
  assumes \<delta>: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "\<Union>set (map trms\<^sub>s\<^sub>t\<^sub>p (S \<cdot>\<^sub>s\<^sub>t \<delta>)) = (\<Union>set (map trms\<^sub>s\<^sub>t\<^sub>p S)) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>"
using assms
proof (induction S)
  case Nil thus ?case by simp
next
  case (Cons x S)
  hence "bvars\<^sub>s\<^sub>t [x] \<inter> subst_domain \<delta> = {}" "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
    unfolding bvars\<^sub>s\<^sub>t_def by force+
  hence *: "subst_domain \<delta> \<inter> set (bvars\<^sub>s\<^sub>t\<^sub>p x) = {}"
           "\<Union>set (map trms\<^sub>s\<^sub>t\<^sub>p (S \<cdot>\<^sub>s\<^sub>t \<delta>)) = \<Union>set (map trms\<^sub>s\<^sub>t\<^sub>p S) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>"
    using Cons.IH(1) bvars\<^sub>s\<^sub>t_singleton[of x] by auto
  hence "trms\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = (trms\<^sub>s\<^sub>t\<^sub>p x) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>"
  proof (cases x)
    case (Inequality X t t') thus ?thesis using rm_vars_apply'[of \<delta> "set X"] * by simp
  qed simp_all
  thus ?case using * subst_all_insert by auto
qed

lemma subst_apply_fv_subset_strand_trm:
  assumes P: "P = fv\<^sub>s\<^sub>t\<^sub>p \<or> P = fv\<^sub>r\<^sub>c\<^sub>v \<or> P = fv\<^sub>s\<^sub>n\<^sub>d \<or> P = fv\<^sub>e\<^sub>q \<or> P = fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q"
  and fv_sub: "fv t \<subseteq> \<Union>set (map P S) \<union> V"
  and \<delta>: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "fv (t \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
using fv_strand_subst[OF P \<delta>] subst_apply_fv_subset[OF fv_sub, of \<delta>] by force

lemma subst_apply_fv_subset_strand_trm2:
  assumes fv_sub: "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V"
  and \<delta>: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "fv (t \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
using fv_strand_subst2[OF \<delta>] subst_apply_fv_subset[OF fv_sub, of \<delta>] by force

lemma subst_apply_fv_subset_strand:
  assumes P: "P = fv\<^sub>s\<^sub>t\<^sub>p \<or> P = fv\<^sub>r\<^sub>c\<^sub>v \<or> P = fv\<^sub>s\<^sub>n\<^sub>d \<or> P = fv\<^sub>e\<^sub>q \<or> P = fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q"
  and P_subset: "P x \<subseteq> \<Union>set (map P S) \<union> V"
  and \<delta>: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
         "set (bvars\<^sub>s\<^sub>t\<^sub>p x) \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
proof (cases x)
  case (Send t)
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>r\<^sub>c\<^sub>v x = {}" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>s\<^sub>n\<^sub>d x = fv t" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>e\<^sub>q x = {}" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = {}" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
    by auto
  hence **: "(P x = fv t \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)) \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})" by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
    hence "fv t \<subseteq> \<Union>set (map P S) \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      unfolding vars\<^sub>s\<^sub>t_def using P subst_apply_fv_subset_strand_trm assms by blast
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)\<close> by force
  }
  ultimately show ?thesis by metis
next
  case (Receive t)
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>r\<^sub>c\<^sub>v x = fv t" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>s\<^sub>n\<^sub>d x = {}" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>e\<^sub>q x = {}" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = {}" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
    by auto
  hence **: "(P x = fv t \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)) \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})" by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
    hence "fv t \<subseteq> \<Union>set (map P S) \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      unfolding vars\<^sub>s\<^sub>t_def using P subst_apply_fv_subset_strand_trm assms by blast
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)\<close> by blast
  }
  ultimately show ?thesis by metis
next
  case (Equality t t')
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t \<union> fv t'" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)"
           "fv\<^sub>r\<^sub>c\<^sub>v x = {}" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>s\<^sub>n\<^sub>d x = {}" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>e\<^sub>q x = fv t \<union> fv t'" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = {}" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
    by auto
  hence **: "(P x = fv t \<union> fv t' \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>))
            \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})"
    by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t \<union> fv t'" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)"
    hence "fv t \<subseteq> \<Union>set (map P S) \<union> V" "fv t' \<subseteq> \<Union>set (map P S) \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
          "fv (t' \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      unfolding vars\<^sub>s\<^sub>t_def using P subst_apply_fv_subset_strand_trm assms by metis+
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)\<close> by blast
  }
  ultimately show ?thesis by metis
next
  case (Inequality X t t')
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t \<union> fv t' - set X" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
           "fv\<^sub>r\<^sub>c\<^sub>v x = {}" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>s\<^sub>n\<^sub>d x = {}" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>e\<^sub>q x = {}" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = (fv t \<union> fv t') - set X"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
    using \<delta>(2) ineq_apply_subst[of \<delta> X t t'] by force+
  hence **: "(P x = fv t \<union> fv t' - set X \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X)
            \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})"
    by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t \<union> fv t' - set X" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
    hence "fv t - set X \<subseteq> \<Union>set (map P S) \<union> V" "fv t' - set X \<subseteq> \<Union>set (map P S) \<union> V"
      using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (V \<union> set X))"
          "fv (t' \<cdot> \<delta>) \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (V \<union> set X))"
      unfolding vars\<^sub>s\<^sub>t_def
      using subst_apply_fv_subset_strand_trm[OF P _ assms(3), of _ "V \<union> set X"] by blast+
    moreover have "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` set X) = set X" using assms(4) Inequality by force
    ultimately have
          "fv (t \<cdot> \<delta>) - set X \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
          "fv (t' \<cdot> \<delta>) - set X \<subseteq> \<Union>set (map P (S \<cdot>\<^sub>s\<^sub>t \<delta>)) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      by force+
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X\<close> by blast
  }
  ultimately show ?thesis by metis
qed

lemma subst_apply_fv_subset_strand2:
  assumes P: "P = fv\<^sub>s\<^sub>t\<^sub>p \<or> P = fv\<^sub>r\<^sub>c\<^sub>v \<or> P = fv\<^sub>s\<^sub>n\<^sub>d \<or> P = fv\<^sub>e\<^sub>q \<or> P = fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q \<or> P = fv_r\<^sub>e\<^sub>q"
  and P_subset: "P x \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V"
  and \<delta>: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
         "set (bvars\<^sub>s\<^sub>t\<^sub>p x) \<inter> (subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) = {}"
  shows "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
proof (cases x)
  case (Send t)
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>r\<^sub>c\<^sub>v x = {}" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>s\<^sub>n\<^sub>d x = fv t" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>e\<^sub>q x = {}" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = {}" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv_r\<^sub>e\<^sub>q x = {}" "fv_r\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
    by auto
  hence **: "(P x = fv t \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)) \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})" by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
    hence "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      using P subst_apply_fv_subset_strand_trm2 assms by blast
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)\<close> by blast
  }
  ultimately show ?thesis by metis
next
  case (Receive t)
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>r\<^sub>c\<^sub>v x = fv t" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
           "fv\<^sub>s\<^sub>n\<^sub>d x = {}" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>e\<^sub>q x = {}" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = {}" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv_r\<^sub>e\<^sub>q x = {}" "fv_r\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
    by auto
  hence **: "(P x = fv t \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)) \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})" by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)"
    hence "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      using P subst_apply_fv_subset_strand_trm2 assms by blast
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>)\<close> by blast
  }
  ultimately show ?thesis by metis
next
  case (Equality t t')
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t \<union> fv t'" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)"
           "fv\<^sub>r\<^sub>c\<^sub>v x = {}" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>s\<^sub>n\<^sub>d x = {}" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>e\<^sub>q x = fv t \<union> fv t'" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = {}" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv_r\<^sub>e\<^sub>q x = fv t'" "fv_r\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t' \<cdot> \<delta>)"
    by auto
  hence **: "(P x = fv t \<union> fv t' \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>))
            \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})
            \<or> (P x = fv t' \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t' \<cdot> \<delta>))"
    by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t \<union> fv t'" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)"
    hence "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
          "fv (t' \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      using P subst_apply_fv_subset_strand_trm2 assms by blast+
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>)\<close> by blast
  }
  moreover
  { assume "P x = fv t'" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t' \<cdot> \<delta>)"
    hence "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" using P_subset by auto
    hence "fv (t' \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      using P subst_apply_fv_subset_strand_trm2 assms by blast+
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t' \<cdot> \<delta>)\<close> by blast
  }
  ultimately show ?thesis by metis
next
  case (Inequality X t t')
  hence *: "fv\<^sub>s\<^sub>t\<^sub>p x = fv t \<union> fv t' - set X" "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
           "fv\<^sub>r\<^sub>c\<^sub>v x = {}" "fv\<^sub>r\<^sub>c\<^sub>v (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>s\<^sub>n\<^sub>d x = {}" "fv\<^sub>s\<^sub>n\<^sub>d (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>e\<^sub>q x = {}" "fv\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
           "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q x = fv t \<union> fv t' - set X" "fv\<^sub>i\<^sub>n\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
           "fv_r\<^sub>e\<^sub>q x = {}" "fv_r\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}"
    using \<delta>(2) ineq_apply_subst[of \<delta> X t t'] by force+
  hence **: "(P x = fv t \<union> fv t' - set X \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X)
            \<or> (P x = {} \<and> P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {})"
    by (metis P)
  moreover
  { assume "P x = {}" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = {}" hence ?thesis by simp }
  moreover
  { assume "P x = fv t \<union> fv t' - set X" "P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X"
    hence "fv t - set X \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" "fv t' - set X \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" using P_subset by auto
    hence "fv (t \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (V \<union> set X))"
          "fv (t' \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (V \<union> set X))"
      unfolding vars\<^sub>s\<^sub>t_def
      using subst_apply_fv_subset_strand_trm2[OF _ assms(3), of _ "V \<union> set X"]
      by blast+
    moreover have "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` set X) = set X" using assms(4) Inequality by force
    ultimately have
          "fv (t \<cdot> \<delta>) - set X \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
          "fv (t' \<cdot> \<delta>) - set X \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
      by fastforce+
    hence ?thesis using \<open>P (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) - set X\<close> by blast
  }
  ultimately show ?thesis by metis
qed

lemma strand_subst_fv_bounded_if_img_bounded:
  assumes "fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t S"
  shows "fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<subseteq> fv\<^sub>s\<^sub>t S"
using subst_sends_strand_fv_to_img[of S \<delta>] assms by blast

lemma strand_fv_subst_subset_if_elim:
  assumes "elim \<delta> v" and "v \<in> fv\<^sub>s\<^sub>t S \<or> bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
  shows "v \<notin> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
proof (cases "v \<in> fv\<^sub>s\<^sub>t S")
  case True thus ?thesis
  proof (induction S)
    case (Cons x S)
    have *: "v \<notin> fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)"
    using assms(1)
    proof (cases x)
      case (Inequality X t t')
      hence "elim (rm_vars (set X) \<delta>) v \<or> v \<in> set X" using assms(1) by blast
      moreover have
          "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> rm_vars (set X) \<delta>) \<union> fv (t' \<cdot> rm_vars (set X) \<delta>) - set X"
        using Inequality by auto
      ultimately show ?thesis using Inequality by blast
    qed simp_all
    moreover have "v \<notin> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)" using Cons.IH
    proof (cases "v \<in> fv\<^sub>s\<^sub>t S")
      case False
      moreover have "v \<notin> fv_subst_range \<delta>" by (simp add: elim_dest''[OF assms(1)]) 
      ultimately show ?thesis by (meson UnE subsetCE subst_sends_strand_fv_to_img)
    qed simp
    ultimately show ?case by auto
  qed simp
next
  case False
  thus ?thesis using assms fv_strand_subst' unfolding elim_def
    by (metis (mono_tags, hide_lams) fv\<^sub>s\<^sub>e\<^sub>t_def imageE mem_simps(8) subst_apply_term.simps(1))
qed

lemma strand_fv_subst_subset_if_elim':
  assumes "elim \<delta> v" "v \<in> fv\<^sub>s\<^sub>t S" "fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t S"
  shows "fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<subset> fv\<^sub>s\<^sub>t S"
using strand_fv_subst_subset_if_elim[OF assms(1)] assms(2)
      strand_subst_fv_bounded_if_img_bounded[OF assms(3)]
  by blast

lemma fv_ik_is_fv_rcv: " fv\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) = \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma fv_ik_subset_fv_st[simp]: "fv\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<subseteq> nonineqvars\<^sub>s\<^sub>t S"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma fv_eqs_rhs_subset_fv_st[simp]: "fv\<^sub>s\<^sub>e\<^sub>t (eqs_rhs\<^sub>s\<^sub>t S) \<subseteq> nonineqvars\<^sub>s\<^sub>t S"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) force+

lemma fv_ik_subset_fv_st'[simp]: "fv\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<subseteq> fv\<^sub>s\<^sub>t S"
by (induct S rule: ik\<^sub>s\<^sub>t.induct) auto

lemma fv_eqs_rhs_subset_fv_st'[simp]: "fv\<^sub>s\<^sub>e\<^sub>t (eqs_rhs\<^sub>s\<^sub>t S) \<subseteq> fv\<^sub>s\<^sub>t S"
by (induct S rule: eqs_rhs\<^sub>s\<^sub>t.induct) auto

lemma strand_step_id_subst[iff]: "x \<cdot>\<^sub>s\<^sub>t\<^sub>p Var = x" by (cases x) simp_all

lemma strand_id_subst[iff]: "S \<cdot>\<^sub>s\<^sub>t Var = S" using strand_step_id_subst by (induct S) auto

lemma strand_subst_vars_union_bound[simp]: "vars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<subseteq> vars\<^sub>s\<^sub>t S \<union> fv_subst_range \<delta>"
proof (induction S)
  case (Cons x S)
  moreover have "vars\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> vars\<^sub>s\<^sub>t\<^sub>p x \<union> fv_subst_range \<delta>" using subst_sends_fv_to_img[of _ \<delta>]
  proof (cases x)
    case (Inequality X t t')
    let ?\<delta> = "rm_vars (set X) \<delta>"
    have "vars\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) = fv (t \<cdot> ?\<delta>) \<union> fv (t' \<cdot> ?\<delta>) \<union> set X" "vars\<^sub>s\<^sub>t\<^sub>p x = fv t \<union> fv t' \<union> set X"
      using Inequality by auto
    moreover have "fv_subst_range ?\<delta> \<subseteq> fv_subst_range \<delta>"
      using rm_vars_img[of "set X" \<delta>] by (auto simp add: subst_domain_def)
    hence "fv (t \<cdot> ?\<delta>) \<subseteq> fv t \<union> fv_subst_range \<delta>" "fv (t' \<cdot> ?\<delta>) \<subseteq> fv t' \<union> fv_subst_range \<delta>"
      using subst_sends_fv_to_img[of _ ?\<delta>] by blast+
    ultimately show ?thesis by auto
  qed auto
  ultimately show ?case by auto
qed simp

lemma strand_vars_split:
  "vars\<^sub>s\<^sub>t (S@S') = vars\<^sub>s\<^sub>t S \<union> vars\<^sub>s\<^sub>t S'" "nonineqvars\<^sub>s\<^sub>t (S@S') = nonineqvars\<^sub>s\<^sub>t S \<union> nonineqvars\<^sub>s\<^sub>t S'"
  "fv\<^sub>s\<^sub>t (S@S') = fv\<^sub>s\<^sub>t S \<union> fv\<^sub>s\<^sub>t S'"
by auto

lemma bvars_subst_ident: "bvars\<^sub>s\<^sub>t S = bvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
unfolding bvars\<^sub>s\<^sub>t_def
by (induct S) (simp_all split: strand_step.splits)

lemma strand_subst_Idem:
  assumes "Idem \<delta>" "subst_domain \<delta> \<union> fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t S" "subst_domain \<theta> \<inter> fv\<^sub>s\<^sub>t S = {}"
          "fv_subst_range \<delta> \<inter> bvars\<^sub>s\<^sub>t S = {}" "fv_subst_range \<theta> \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "(S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t \<theta> = (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
  and   "(S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t (\<theta> \<circ>\<^sub>s \<delta>) = (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
proof -
  from assms(2,3) have "fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) \<inter> subst_domain \<theta> = {}"
    using subst_sends_strand_fv_to_img[of S \<delta>] by blast
  thus "(S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t \<theta> = (S \<cdot>\<^sub>s\<^sub>t \<delta>)" by blast
  thus "(S \<cdot>\<^sub>s\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>t (\<theta> \<circ>\<^sub>s \<delta>) = (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
    by (metis assms(1,4,5) bvars_subst_ident strand_subst_comp Idem_def)
qed

lemma strand_subst_img_bound:
  assumes "subst_domain \<delta> \<union> fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t S" "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
proof -
  have "subst_domain \<delta> \<subseteq> \<Union>set (map fv\<^sub>s\<^sub>t\<^sub>p S)" by (metis (no_types) fv\<^sub>s\<^sub>t_def Un_subset_iff assms(1))
  thus ?thesis unfolding fv\<^sub>s\<^sub>t_def
    by (metis subst_range.simps fv_set_mono fv_strand_subst Int_commute assms(2) image_Un le_iff_sup)
qed

lemma strand_subst_img_bound':
  assumes "subst_domain \<delta> \<union> fv_subst_range \<delta> \<subseteq> vars\<^sub>s\<^sub>t S" "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "fv_subst_range \<delta> \<subseteq> vars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
proof -
  have "(subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` subst_domain \<delta>)) \<inter> vars\<^sub>s\<^sub>t S = subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` subst_domain \<delta>)"
    using assms(1) by (metis inf.absorb_iff1 subst_range.simps)
  hence "fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
    using vars_snd_rcv_strand fv_snd_rcv_strand assms(2) strand_subst_img_bound
    by (metis (no_types) inf_le2 inf_sup_distrib1 subst_range.simps sup_bot.right_neutral)
  thus "fv_subst_range \<delta> \<subseteq> vars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)" by (metis fv_snd_rcv_strand le_supI1 vars_snd_rcv_strand)
qed

lemma strand_subst_all_fv_subset:
  assumes "fv t \<subseteq> fv\<^sub>s\<^sub>t S" "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "fv (t \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
using assms by (metis fv_strand_subst' Int_commute subst_apply_fv_subset)

lemma strand_subst_not_dom_fixed: "\<lbrakk>v \<in> fv\<^sub>s\<^sub>t S; v \<notin> subst_domain \<delta>\<rbrakk> \<Longrightarrow> v \<in> fv\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
proof (induction S)
  case (Cons x S') show ?case
  proof (cases "v \<in> fv\<^sub>s\<^sub>t S'")
    case True thus ?thesis using Cons.IH[OF _ Cons.prems(2)] by auto
  next
    case False
    hence *: "v \<in> fv\<^sub>s\<^sub>t\<^sub>p x" using Cons.prems(1) by simp
    hence "v \<in> fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)"
      using Cons.prems(2) subst_not_dom_fixed
      by (cases x) (force simp add: subst_domain_def)+
    thus ?thesis by auto
  qed
qed simp

lemma strand_vars_unfold: "v \<in> vars\<^sub>s\<^sub>t S \<Longrightarrow> \<exists>S' x S''. S = S'@x#S'' \<and> v \<in> vars\<^sub>s\<^sub>t\<^sub>p x"
proof (induction S)
  case (Cons x S) thus ?case
  proof (cases "v \<in> vars\<^sub>s\<^sub>t\<^sub>p x")
    case True thus ?thesis by blast
  next
    case False
    hence "v \<in> vars\<^sub>s\<^sub>t S" using Cons.prems by auto
    thus ?thesis using Cons.IH by (metis append_Cons)
  qed
qed simp

lemma strand_fv_unfold: "v \<in> fv\<^sub>s\<^sub>t S \<Longrightarrow> \<exists>S' x S''. S = S'@x#S'' \<and> v \<in> fv\<^sub>s\<^sub>t\<^sub>p x"
proof (induction S)
  case (Cons x S) thus ?case
  proof (cases "v \<in> fv\<^sub>s\<^sub>t\<^sub>p x")
    case True thus ?thesis by blast
  next
    case False
    hence "v \<in> fv\<^sub>s\<^sub>t S" using Cons.prems by auto
    thus ?thesis using Cons.IH by (metis append_Cons)
  qed
qed simp

lemma (in intruder_model) wf_fv_strand_if_in_trp_ik_union:
  "\<lbrakk>t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S); wf\<^sub>s\<^sub>t V S\<rbrakk>  \<Longrightarrow> fv t \<subseteq> fv\<^sub>s\<^sub>t S \<union> V"
proof (induction S arbitrary: V rule: ik\<^sub>s\<^sub>t_induct)
  case (ConsRcv t' S)
  hence "fv t' \<subseteq> V" "wf\<^sub>s\<^sub>t V S" by auto
  thus ?case using ConsRcv.IH ConsRcv.prems
  proof (cases "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)")
    case False
    hence "t \<sqsubseteq>\<^sub>t\<^sub>r\<^sub>p t'" using \<open>t \<in> \<Union>(trp ` ik\<^sub>s\<^sub>t (receive\<langle>t'\<rangle>#S))\<close> by simp
    thus ?thesis using \<open>fv t' \<subseteq> V\<close> trpD[THEN subtermeq_vars_subset] by auto
  qed auto
qed auto

lemma (in intruder_model) subterm_if_in_strand_ik:
  "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<Longrightarrow> \<exists>t'. receive\<langle>t'\<rangle> \<in> set S \<and> t \<sqsubseteq> t'"
proof (induction S rule: ik\<^sub>s\<^sub>t_induct)
  case (ConsRcv t' S) thus ?case by (cases "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)", auto dest!: trpD)
qed simp_all

lemma (in intruder_model) fv_subset_if_in_strand_ik:
  "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<Longrightarrow> fv t \<subseteq> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)"
proof -
  assume "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)"
  then obtain t' where "receive\<langle>t'\<rangle> \<in> set S" "t \<sqsubseteq> t'" by (metis subterm_if_in_strand_ik)
  hence "fv t \<subseteq> fv t'" by (simp add: subtermeq_vars_subset)
  thus ?thesis using in_strand_fv_subset_rcv[OF \<open>receive\<langle>t'\<rangle> \<in> set S\<close>] by auto
qed

lemma (in intruder_model) fv_subset_if_in_strand_ik':
  "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<Longrightarrow> fv t \<subseteq> fv\<^sub>s\<^sub>t S"
using fv_subset_if_in_strand_ik[of t S] fv_snd_rcv_strand_subset(2)[of S] by blast

lemma (in intruder_model) vars_subset_if_in_strand_ik2:
  "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S) \<Longrightarrow> fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S"
using fv_subset_if_in_strand_ik[of t S] vars_snd_rcv_strand_subset2(2)[of S] by blast


subsection {* Lemmas: Simple Strands *}
lemma simple_Cons[dest]: "simple (s#S) \<Longrightarrow> simple S"
unfolding simple_def by auto

lemma simple_split[dest]:
  assumes "simple (S@S')"
  shows "simple S" "simple S'"
using assms unfolding simple_def by auto

lemma simple_append[intro]: "\<lbrakk>simple S; simple S'\<rbrakk> \<Longrightarrow> simple (S@S')"
unfolding simple_def by auto

lemma simple_append_sym[sym]: "simple (S@S') \<Longrightarrow> simple (S'@S)" by auto

lemma not_simple_if_snd_fun: "(\<exists>S' S'' f X. S = S'@send\<langle>Fun f X\<rangle>#S'') \<Longrightarrow> \<not>simple S"
unfolding simple_def by auto

lemma not_list_all_elim: "\<not>list_all P A \<Longrightarrow> \<exists>B x C. A = B@x#C \<and> \<not>P x \<and> list_all P B"
proof (induction A rule: List.rev_induct)
  case (snoc a A)
  show ?case
  proof (cases "list_all P A")
    case True
    thus ?thesis using snoc.prems by auto
  next
    case False
    then obtain B x C where "A = B@x#C" "\<not>P x" "list_all P B" using snoc.IH[OF False] by auto
    thus ?thesis by auto
  qed
qed simp

lemma not_simple\<^sub>s\<^sub>t\<^sub>p_elim:
  assumes "\<not>simple\<^sub>s\<^sub>t\<^sub>p x"
  shows "(\<exists>f T. x = send\<langle>Fun f T\<rangle>) \<or> 
         (\<exists>t t'. x = \<langle>t \<doteq> t'\<rangle>) \<or>
         (\<exists>X t t'. x = \<nexists>X\<langle>t \<doteq> t'\<rangle> \<and> \<not>(\<exists>\<I>. ineq_model \<I> X t t'))"
using assms by (cases x) (fastforce elim: simple\<^sub>s\<^sub>t\<^sub>p.elims)+

lemma not_simple_elim:
  assumes "\<not>simple S"
  shows "(\<exists>A B f T. S = A@send\<langle>Fun f T\<rangle>#B \<and> simple A) \<or> 
         (\<exists>A B t t'. S = A@\<langle>t \<doteq> t'\<rangle>#B \<and> simple A) \<or>
         (\<exists>A B X t t'. S = A@\<nexists>X\<langle>t \<doteq> t'\<rangle>#B \<and> \<not>(\<exists>\<I>. ineq_model \<I> X t t'))"
by (metis assms not_list_all_elim not_simple\<^sub>s\<^sub>t\<^sub>p_elim simple_def)

lemma simple_fun_prefix_unique:
  assumes "A = S@send\<langle>Fun f X\<rangle>#S'" "simple S"
  shows "\<forall>T g Y T'. A = T@send\<langle>Fun g Y\<rangle>#T' \<and> simple T \<longrightarrow> S = T \<and> f = g \<and> X = Y \<and> S' = T'"
proof -
  { fix T g Y T' assume *: "A = T@send\<langle>Fun g Y\<rangle>#T'" "simple T"
    { assume "length S < length T" hence False using assms *
        by (metis id_take_nth_drop not_simple_if_snd_fun nth_append nth_append_length)
    }
    moreover
    { assume "length S > length T" hence False using assms *
        by (metis id_take_nth_drop not_simple_if_snd_fun nth_append nth_append_length)
    }
    ultimately have "S = T" using assms * by (meson List.append_eq_append_conv linorder_neqE_nat)
  }
  thus ?thesis using assms(1) by blast
qed

lemma simple_snd_is_var: "\<lbrakk>send\<langle>t\<rangle> \<in> set S; simple S\<rbrakk> \<Longrightarrow> \<exists>v. t = Var v"
unfolding simple_def
by (metis list_all_append list_all_simps(1) simple\<^sub>s\<^sub>t\<^sub>p.elims(2) split_list_first
          strand_step.distinct(1) strand_step.distinct(5) strand_step.inject(1)) 


subsection {* Lemmas: Strand Measure *}
lemma measure\<^sub>s\<^sub>t_wellfounded: "wf measure\<^sub>s\<^sub>t" unfolding measure\<^sub>s\<^sub>t_def by simp

lemma strand_size_append[iff]: "size\<^sub>s\<^sub>t (S@S') = size\<^sub>s\<^sub>t S + size\<^sub>s\<^sub>t S'"
by (induct S) auto

lemma strand_size_map_fun_lt[simp]:
  "size\<^sub>s\<^sub>t (map Send X) < size (Fun f X)"
  "size\<^sub>s\<^sub>t (map Send X) < size\<^sub>s\<^sub>t [send\<langle>Fun f X\<rangle>]"
  "size\<^sub>s\<^sub>t (map Send X) < size\<^sub>s\<^sub>t [receive\<langle>Fun f X\<rangle>]"
by (induct X) auto

lemma strand_size_rm_fun_lt[simp]:
  "size\<^sub>s\<^sub>t (S@S') < size\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')"
  "size\<^sub>s\<^sub>t (S@S') < size\<^sub>s\<^sub>t (S@receive\<langle>Fun f X\<rangle>#S')"
by (induct S) auto

lemma strand_fv_card_map_fun_eq:
  "card (fv\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')) = card (fv\<^sub>s\<^sub>t (S@(map Send X)@S'))"
proof -
  have "fv\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S') = fv\<^sub>s\<^sub>t (S@(map Send X)@S')" by auto
  thus ?thesis by simp
qed

lemma strand_fv_card_rm_fun_le[simp]: "card (fv\<^sub>s\<^sub>t (S@S')) \<le> card (fv\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S'))"
by (force intro: card_mono)

lemma strand_fv_card_rm_eq_le[simp]: "card (fv\<^sub>s\<^sub>t (S@S')) \<le> card (fv\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S'))"
by (force intro: card_mono)


subsection {* Lemmas: Well-formed Strands *}
lemma wf_prefix[dest]: "wf\<^sub>s\<^sub>t V (S@S') \<Longrightarrow> wf\<^sub>s\<^sub>t V S"
by (induct S rule: wf\<^sub>s\<^sub>t.induct) auto

lemma wf_vars_mono[simp]: "wf\<^sub>s\<^sub>t V S \<Longrightarrow> wf\<^sub>s\<^sub>t (V \<union> W) S"
proof (induction S arbitrary: V)
  case (Cons x S) thus ?case
  proof (cases x)
    case (Send t)
    hence "wf\<^sub>s\<^sub>t (V \<union> fv t \<union> W) S" using Cons.prems(1) Cons.IH by simp
    thus ?thesis using Send by (simp add: sup_commute sup_left_commute)
  next
    case (Equality t t')
    hence "wf\<^sub>s\<^sub>t (V \<union> fv t \<union> W) S" "fv t' \<subseteq> V \<union> W" using Cons.prems(1) Cons.IH by auto
    thus ?thesis using Equality by (simp add: sup_commute sup_left_commute)
  qed auto
qed simp

lemma wf\<^sub>s\<^sub>tI[intro]: "nonineqvars\<^sub>s\<^sub>t S \<subseteq> V \<Longrightarrow> wf\<^sub>s\<^sub>t V S"
proof (induction S)
  case (Cons x S) thus ?case
  proof (cases x)
    case (Send t)
    hence "wf\<^sub>s\<^sub>t V S" "V \<union> fv t = V" using Cons by auto
    thus ?thesis using Send by simp
  next
    case (Equality t t')
    hence "wf\<^sub>s\<^sub>t V S" "fv t' \<subseteq> V" using Cons by auto
    thus ?thesis using wf_vars_mono Equality by simp
  qed simp_all
qed simp

lemma wf_append_suffix: "wf\<^sub>s\<^sub>t V S \<Longrightarrow> nonineqvars\<^sub>s\<^sub>t S' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@S')"
proof (induction V S rule: wf\<^sub>s\<^sub>t_induct)
  case (ConsSnd V t S)
  hence *: "wf\<^sub>s\<^sub>t (V \<union> fv t) S" by simp_all
  hence "nonineqvars\<^sub>s\<^sub>t S' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> (V \<union> fv t)" using ConsSnd.prems(2) by fastforce
  thus ?case using ConsSnd.IH * by simp
next
  case (ConsRcv V t S)
  hence *: "fv t \<subseteq> V" "wf\<^sub>s\<^sub>t V S" by simp_all
  hence "nonineqvars\<^sub>s\<^sub>t S' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" using ConsRcv.prems(2) by fastforce
  thus ?case using ConsRcv.IH * by simp
next
  case (ConsEq V t t' S)
  hence *: "fv t' \<subseteq> V" "wf\<^sub>s\<^sub>t (V \<union> fv t) S" by simp_all
  moreover have "vars\<^sub>s\<^sub>t\<^sub>p (\<langle>t \<doteq> t'\<rangle>) = fv t \<union> fv t'" by simp
  moreover have "nonineqvars\<^sub>s\<^sub>t (\<langle>t \<doteq> t'\<rangle>#S) = fv t \<union> fv t' \<union> nonineqvars\<^sub>s\<^sub>t S" by auto
  ultimately have "nonineqvars\<^sub>s\<^sub>t S' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> (V \<union> fv t)" using ConsEq.prems(2) by blast
  thus ?case using ConsEq.IH * by simp
qed (simp_all add: wf\<^sub>s\<^sub>tI)

lemma wf_send_compose: "wf\<^sub>s\<^sub>t V (S@(map Send X)@S') = wf\<^sub>s\<^sub>t V (S@send\<langle>Fun f X\<rangle>#S')"
proof (induction S arbitrary: V)
  case Nil thus ?case 
  proof (induction X arbitrary: V)
    case (Cons y Y) thus ?case by (simp add: sup_assoc)
  qed simp
next
  case (Cons s S) thus ?case by (cases s) simp_all
qed

lemma wf_snd_append[iff]: "wf\<^sub>s\<^sub>t V (S@[send\<langle>t\<rangle>]) = wf\<^sub>s\<^sub>t V S"
by (induct S rule: wf\<^sub>s\<^sub>t.induct) simp_all

lemma wf_snd_append': "wf\<^sub>s\<^sub>t V S \<Longrightarrow> wf\<^sub>s\<^sub>t V (send\<langle>t\<rangle>#S)"
by simp

lemma wf_rcv_append[dest]: "wf\<^sub>s\<^sub>t V (S@receive\<langle>t\<rangle>#S') \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@S')"
by (induct S rule: wf\<^sub>s\<^sub>t.induct) simp_all

lemma wf_rcv_append'[intro]: "\<lbrakk>wf\<^sub>s\<^sub>t V (S@S'); fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V\<rbrakk> \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@receive\<langle>t\<rangle>#S')"
proof (induction S rule: wf\<^sub>s\<^sub>t_induct)
  case (ConsRcv V t' S)
  hence "wf\<^sub>s\<^sub>t V (S@S')" "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" by fastforce+
  thus ?case using ConsRcv by auto
next
  case (ConsEq V t' t'' S)
  hence "fv t'' \<subseteq> V" by simp
  moreover have "nonineqvars\<^sub>s\<^sub>t (\<langle>t' \<doteq> t''\<rangle>#S) = fv t' \<union> fv t'' \<union> nonineqvars\<^sub>s\<^sub>t S" by auto
  ultimately have "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> (V \<union> fv t')" using ConsEq.prems(2) by blast
  thus ?case using ConsEq by auto
qed auto

lemma wf_rcv_append''[intro]: "\<lbrakk>wf\<^sub>s\<^sub>t V S; fv t \<subseteq> \<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)\<rbrakk> \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@[receive\<langle>t\<rangle>])"
apply (induct S, simp)
by (metis vars_snd_rcv_strand_subset2(1) append_Nil2 le_supI1 order_trans wf_rcv_append')

lemma wf_rcv_append'''[intro]: "\<lbrakk>wf\<^sub>s\<^sub>t V S; fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V\<rbrakk> \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@[receive\<langle>t\<rangle>])"
by (simp add: wf_rcv_append'[of _ _ "[]"])

lemma wf_eq_append[dest]: "wf\<^sub>s\<^sub>t V (S@\<langle>t \<doteq> t'\<rangle>#S') \<Longrightarrow> fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@S')"
proof (induction S rule: wf\<^sub>s\<^sub>t_induct)
  case (Nil V)
  hence "V \<union> fv t = V" "wf\<^sub>s\<^sub>t (V \<union> fv t) S'" by auto
  thus ?case by auto
next
  case (ConsRcv V u S)
  hence "wf\<^sub>s\<^sub>t V (S @ \<langle>t \<doteq> t'\<rangle> # S')" "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" "fv u \<subseteq> V" by fastforce+
  hence "wf\<^sub>s\<^sub>t V (S@S')" using ConsRcv.IH by auto
  thus ?case using \<open>fv u \<subseteq> V\<close> by simp
next
  case (ConsEq V u u' S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv u) (S@\<langle>t \<doteq> t'\<rangle>#S')" "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> (V \<union> fv u)" "fv u' \<subseteq> V" by auto
  hence "wf\<^sub>s\<^sub>t (V \<union> fv u) (S@S')" using ConsEq.IH by auto
  thus ?case using \<open>fv u' \<subseteq> V\<close> by simp
qed auto

lemma wf_eq_append'[intro]: "\<lbrakk>wf\<^sub>s\<^sub>t V (S@S'); fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V\<rbrakk> \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@\<langle>t \<doteq> t'\<rangle>#S')"
proof (induction S rule: wf\<^sub>s\<^sub>t_induct)
  case (ConsEq V u u' S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv u) (S@S')" "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V \<union> fv u" by auto+
  thus ?case using ConsEq by auto
next
  case (ConsRcv V u S)
  hence "wf\<^sub>s\<^sub>t V (S@S')" "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V" by fastforce+
  thus ?case using ConsRcv by auto
next
  case (ConsSnd V u S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv u) (S@S')" "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> (V \<union> fv u)" by fastforce+
  thus ?case using ConsSnd by auto
qed auto

lemma wf_eq_append''[intro]: "\<lbrakk>wf\<^sub>s\<^sub>t V (S@S'); fv t' \<subseteq> wfvarsoccs\<^sub>s\<^sub>t S \<union> V\<rbrakk> \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@[\<langle>t \<doteq> t'\<rangle>]@S')"
proof (induction S rule: wf\<^sub>s\<^sub>t_induct)
  case (ConsEq V u u' S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv u) (S@S')" "fv t' \<subseteq> wfvarsoccs\<^sub>s\<^sub>t S \<union> V \<union> fv u" by fastforce+
  thus ?case using ConsEq by auto
next
  case (ConsRcv V u S)
  hence "wf\<^sub>s\<^sub>t V (S@S')" "fv t' \<subseteq> wfvarsoccs\<^sub>s\<^sub>t S \<union> V" by fastforce+
  thus ?case using ConsRcv by auto
next
  case (ConsSnd V u S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv u) (S@S')" "fv t' \<subseteq> wfvarsoccs\<^sub>s\<^sub>t S \<union> (V \<union> fv u)" by auto
  thus ?case using ConsSnd by auto
qed auto

lemma wf_eq_append'''[intro]: "\<lbrakk>wf\<^sub>s\<^sub>t V S; fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S \<union> V\<rbrakk> \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@[\<langle>t \<doteq> t'\<rangle>])"
by (simp add: wf_eq_append'[of _ _ "[]"])

lemma wf_ineq_append[dest]: "wf\<^sub>s\<^sub>t V (S@\<nexists>X\<langle>t \<doteq> t'\<rangle>#S') \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@S')"
by (induct S rule: wf\<^sub>s\<^sub>t.induct) simp_all

lemma wf_ineq_append'[intro]: "wf\<^sub>s\<^sub>t V (S@S') \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@\<nexists>X\<langle>t \<doteq> t'\<rangle>#S')"
by (induct S rule: wf\<^sub>s\<^sub>t.induct) auto

lemma wf_ineq_append''[intro]: "wf\<^sub>s\<^sub>t V S \<Longrightarrow> wf\<^sub>s\<^sub>t V (S@[\<nexists>X\<langle>t \<doteq> t'\<rangle>])"
by (induct S rule: wf\<^sub>s\<^sub>t.induct) auto

lemma wf_rcv_fv_single[elim]: "wf\<^sub>s\<^sub>t V (receive\<langle>t\<rangle>#S') \<Longrightarrow> fv t \<subseteq> V"
by simp

lemma wf_append_exec: "wf\<^sub>s\<^sub>t V (S@S') \<Longrightarrow> wf\<^sub>s\<^sub>t (V \<union> wfvarsoccs\<^sub>s\<^sub>t S) S'"
proof (induction S arbitrary: V)
  case (Cons x S V) thus ?case
  proof (cases x)
    case (Send t)
    hence "wf\<^sub>s\<^sub>t (V \<union> fv t \<union> wfvarsoccs\<^sub>s\<^sub>t S) S'" using Cons.prems Cons.IH by simp
    thus ?thesis using Send by (auto simp add: sup_assoc)
  next
    case (Equality t t')
    hence "wf\<^sub>s\<^sub>t (V \<union> fv t \<union> wfvarsoccs\<^sub>s\<^sub>t S) S'" using Cons.prems Cons.IH by simp
    thus ?thesis using Equality by (auto simp add: sup_assoc)
  qed auto
qed simp

lemma wf_rcv_fv: "wf\<^sub>s\<^sub>t V (S@receive\<langle>t\<rangle>#S') \<Longrightarrow> fv t \<subseteq> wfvarsoccs\<^sub>s\<^sub>t S \<union> V"
proof (induction S arbitrary: V)
  case (Cons x S V) thus ?case by (cases x) auto
qed simp

lemma wf_eq_fv: "wf\<^sub>s\<^sub>t V (S@\<langle>t \<doteq> t'\<rangle>#S') \<Longrightarrow> fv t' \<subseteq> wfvarsoccs\<^sub>s\<^sub>t S \<union> V"
proof (induction S arbitrary: V)
  case (Cons x S V) thus ?case by (cases x) auto
qed simp

lemma wf_simple_fv_occurrence:
  assumes "wf\<^sub>s\<^sub>t {} S" "simple S" "v \<in> nonineqvars\<^sub>s\<^sub>t S"
  shows "\<exists>S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>Var v\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> v \<notin> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e"
using assms
proof (induction S rule: List.rev_induct)
  case (snoc x S)
  from \<open>wf\<^sub>s\<^sub>t {} (S@[x])\<close> have "wf\<^sub>s\<^sub>t {} S" "wf\<^sub>s\<^sub>t (nonineqvars\<^sub>s\<^sub>t S) [x]"
    using wf_append_exec[THEN wf_vars_mono, of "{}" S "[x]" "nonineqvars\<^sub>s\<^sub>t S - wfvarsoccs\<^sub>s\<^sub>t S"]
          vars_snd_rcv_strand_subset2(4)[of S] Diff_partition[of "wfvarsoccs\<^sub>s\<^sub>t S" "nonineqvars\<^sub>s\<^sub>t S"]
    by auto
  from \<open>simple (S@[x])\<close> have "simple S" "simple\<^sub>s\<^sub>t\<^sub>p x" unfolding simple_def by auto

  show ?case
  proof (cases "v \<in> nonineqvars\<^sub>s\<^sub>t S")
    case True thus ?thesis using snoc.IH[OF \<open>wf\<^sub>s\<^sub>t {} S\<close> \<open>simple S\<close>] by fastforce
  next
    case False
    show ?thesis
    proof (cases x)
      case (Receive t)
      hence "fv t \<subseteq> nonineqvars\<^sub>s\<^sub>t S" using \<open>wf\<^sub>s\<^sub>t (nonineqvars\<^sub>s\<^sub>t S) [x]\<close> by simp
      hence "v \<in> nonineqvars\<^sub>s\<^sub>t S" using \<open>v \<in> nonineqvars\<^sub>s\<^sub>t (S@[x])\<close> \<open>x = receive\<langle>t\<rangle>\<close> by auto
      thus ?thesis using \<open>x = receive\<langle>t\<rangle>\<close> snoc.IH[OF \<open>wf\<^sub>s\<^sub>t {} S\<close> \<open>simple S\<close>] by fastforce
    next
      case (Send t)
      hence "v \<in> vars\<^sub>s\<^sub>t\<^sub>p x" using \<open>v \<in> nonineqvars\<^sub>s\<^sub>t (S@[x])\<close> False by auto
      from Send obtain w where "t = Var w" using \<open>simple\<^sub>s\<^sub>t\<^sub>p x\<close> by (cases t) simp_all
      
      hence "v = w" using \<open>x = send\<langle>t\<rangle>\<close> \<open>v \<in> vars\<^sub>s\<^sub>t\<^sub>p x\<close> by simp
      thus ?thesis using \<open>x = send\<langle>t\<rangle>\<close> \<open>v \<notin> nonineqvars\<^sub>s\<^sub>t S\<close> \<open>t = Var w\<close> by auto
    next
      case (Equality t t') thus ?thesis using snoc.prems(2) unfolding simple_def by auto
    next
      case (Inequality t t') thus ?thesis using False snoc.prems(3) by auto
    qed
  qed
qed simp

lemma (in intruder_model) Unifier_strand_fv_subset:
  assumes g_in_ik: "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)"
  and \<delta>: "Unifier \<delta> (Fun f X) t"
  and disj: "bvars\<^sub>s\<^sub>t S \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
  shows "fv (Fun f X \<cdot> \<delta>) \<subseteq> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v (S \<cdot>\<^sub>s\<^sub>t \<delta>))"
proof -
  have "fv t \<subseteq> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v S)" using fv_subset_if_in_strand_ik[OF g_in_ik] by simp
  hence "fv (t \<cdot> \<delta>) \<subseteq> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v (S \<cdot>\<^sub>s\<^sub>t \<delta>))" using fv_strand_subst
    by (metis (no_types) disj fv_strand_subst subst_apply_fv_subset)
  thus ?thesis using \<delta> by fastforce
qed

lemma wf\<^sub>s\<^sub>t_induct'[consumes 1, case_names Nil ConsSnd ConsRcv ConsEq ConsIneq]:
  fixes S::"('a,'b) strand"
  assumes "wf\<^sub>s\<^sub>t V S"
          "P []"
          "\<And>t S. \<lbrakk>wf\<^sub>s\<^sub>t V S; P S\<rbrakk> \<Longrightarrow> P (S@[send\<langle>t\<rangle>])"
          "\<And>t S. \<lbrakk>wf\<^sub>s\<^sub>t V S; P S; fv t \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t S\<rbrakk> \<Longrightarrow> P (S@[receive\<langle>t\<rangle>])"
          "\<And>t t' S. \<lbrakk>wf\<^sub>s\<^sub>t V S; P S; fv t' \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t S\<rbrakk> \<Longrightarrow> P (S@[\<langle>t \<doteq> t'\<rangle>])"
          "\<And>X t t' S. \<lbrakk>wf\<^sub>s\<^sub>t V S; P S\<rbrakk> \<Longrightarrow> P (S@[\<nexists>X\<langle>t \<doteq> t'\<rangle>])"
  shows "P S"
using assms
proof (induction S rule: List.rev_induct)
  case (snoc x S)
  hence *: "wf\<^sub>s\<^sub>t V S" "wf\<^sub>s\<^sub>t (V \<union> wfvarsoccs\<^sub>s\<^sub>t S) [x]"
    by (metis wf_prefix, metis wf_append_exec)
  show ?case
  proof (cases x)
    case (Send t) thus ?thesis using *(1) snoc.prems snoc.IH by auto
  next
    case (Equality t t') thus ?thesis using *(2) snoc.prems snoc.IH by auto
  next
    case (Inequality t t') thus ?thesis using *(2) snoc.prems snoc.IH by auto
  next
    case (Receive t)
    hence "P S" using snoc.prems snoc.IH[OF *(1) \<open>P []\<close>] by auto
    thus ?thesis using \<open>x = receive\<langle>t\<rangle>\<close> snoc.prems(4)[OF *(1), of t] *(2) by auto
  qed
qed simp

lemma (in intruder_model) wf_subst_apply:
  "wf\<^sub>s\<^sub>t V S \<Longrightarrow> wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) (S \<cdot>\<^sub>s\<^sub>t \<delta>)"
proof (induction S arbitrary: V rule: wf\<^sub>s\<^sub>t_induct)
  case (ConsRcv V t S)
  hence "wf\<^sub>s\<^sub>t V (receive\<langle>t\<rangle>#S)" "fv t \<subseteq> V" by simp_all
  hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) (S \<cdot>\<^sub>s\<^sub>t \<delta>)" "fv (t \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
    using ConsRcv.IH subst_apply_fv_subset by simp_all
  thus ?case by simp
next
  case (ConsSnd V t S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv t) S" by simp
  hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (V \<union> fv t))) (S \<cdot>\<^sub>s\<^sub>t \<delta>)" using ConsSnd.IH by metis
  hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> (fv (t \<cdot> \<delta>))) (S \<cdot>\<^sub>s\<^sub>t \<delta>)" using subst_apply_fv_union by metis
  thus ?case by simp
next
  case (ConsEq V t t' S)
  hence "wf\<^sub>s\<^sub>t (V \<union> fv t) (receive\<langle>t\<rangle>#S)" "fv t' \<subseteq> V" by simp_all
  hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` (V \<union> fv t))) (S \<cdot>\<^sub>s\<^sub>t \<delta>)" and *: "fv (t' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)"
    using ConsEq.IH subst_apply_fv_subset by force+
  hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> (fv (t \<cdot> \<delta>))) (S \<cdot>\<^sub>s\<^sub>t \<delta>)" using subst_apply_fv_union by metis
  thus ?case using * by simp
qed simp_all

lemma (in intruder_model) wf_unify:
  assumes wf: "wf\<^sub>s\<^sub>t V (S@send\<langle>Fun f X\<rangle>#S')"
  and g_in_ik: "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)"
  and \<delta>: "Unifier \<delta> (Fun f X) t"
  and disj: "bvars\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S') \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
  shows "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
using assms
proof (induction S' arbitrary: V rule: List.rev_induct)
  case (snoc x S' V)

  have fun_fv_bound: "fv (Fun f X \<cdot> \<delta>) \<subseteq> \<Union>set (map fv\<^sub>r\<^sub>c\<^sub>v (S \<cdot>\<^sub>s\<^sub>t \<delta>))"
    using snoc.prems(4) bvars\<^sub>s\<^sub>t_split Unifier_strand_fv_subset[OF g_in_ik \<delta>] by auto
  hence "fv (Fun f X \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>))" using fv_ik_is_fv_rcv by metis
  hence "fv (Fun f X \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>)" using fv_ik_subset_fv_st[of "S \<cdot>\<^sub>s\<^sub>t \<delta>"] by blast
  hence *: "fv ((Fun f X) \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)" by fastforce

  from snoc.prems(1) have "wf\<^sub>s\<^sub>t V (S@send\<langle>Fun f X\<rangle>#S')"
    using wf_prefix[of V "S@send\<langle>Fun f X\<rangle>#S'" "[x]"] by simp
  hence **: "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
    using snoc.IH[OF _ snoc.prems(2,3)] snoc.prems(4) by auto

  from snoc.prems(1) have ***: "wf\<^sub>s\<^sub>t (V \<union> wfvarsoccs\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')) [x]"
    using wf_append_exec[of V "(S@send\<langle>Fun f X\<rangle>#S')" "[x]"] by simp

  from snoc.prems(4) have disj':
      "bvars\<^sub>s\<^sub>t (S@S') \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
      "set (bvars\<^sub>s\<^sub>t\<^sub>p x) \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
    by auto

  show ?case
  proof (cases x)
    case (Send t)
    thus ?thesis using wf_snd_append[of "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)" "(S@S') \<cdot>\<^sub>s\<^sub>t \<delta>"] ** by auto
  next
    case (Receive t)
    hence "fv\<^sub>s\<^sub>t\<^sub>p x \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')" using *** by auto
    hence "fv\<^sub>s\<^sub>t\<^sub>p x \<subseteq> V \<union> nonineqvars\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')"
      using vars_snd_rcv_strand_subset2(4)[of "S@send\<langle>Fun f X\<rangle>#S'"] by blast
    hence "fv\<^sub>s\<^sub>t\<^sub>p x \<subseteq> V \<union> fv (Fun f X) \<union> nonineqvars\<^sub>s\<^sub>t (S@S')" by auto
    hence "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> fv ((Fun f X) \<cdot> \<delta>) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
      by (metis (no_types) inf_sup_aci(5) subst_apply_fv_subset_strand2 subst_apply_fv_union disj')
    hence "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)" using * by blast
    hence "fv (t \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) " using \<open>x = receive\<langle>t\<rangle>\<close> by auto
    hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) (((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)@[receive\<langle>t \<cdot> \<delta>\<rangle>])"
      using wf_rcv_append'''[OF **, of "t \<cdot> \<delta>"] by metis
    thus ?thesis using \<open>x = receive\<langle>t\<rangle>\<close> by auto
  next
    case (Equality s s')
    hence "fv s' \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')" using *** by auto
    hence "fv s' \<subseteq> V \<union> nonineqvars\<^sub>s\<^sub>t (S@send\<langle>Fun f X\<rangle>#S')"
      using vars_snd_rcv_strand_subset2(4)[of "S@send\<langle>Fun f X\<rangle>#S'"] by blast
    hence "fv s' \<subseteq> V \<union> fv (Fun f X) \<union> nonineqvars\<^sub>s\<^sub>t (S@S')" by auto
    moreover have "fv s' = fv_r\<^sub>e\<^sub>q x" "fv (s' \<cdot> \<delta>) = fv_r\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)" using Equality by simp_all
    ultimately have "fv (s' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> fv (Fun f X \<cdot> \<delta>) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
      using subst_apply_fv_subset_strand2[of fv_r\<^sub>e\<^sub>q x]
      by (metis (no_types) inf_sup_aci(5) subst_apply_fv_union disj')
    hence "fv (s' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)" using * by blast
    hence "fv (s' \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)" using \<open>x = \<langle>s \<doteq> s'\<rangle>\<close> by auto
    hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) (((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)@[\<langle>(s \<cdot> \<delta>) \<doteq> (s' \<cdot> \<delta>)\<rangle>])"
      using wf_eq_append'''[OF **] by metis
    thus ?thesis using \<open>x = \<langle>s \<doteq> s'\<rangle>\<close> by auto
  next
    case (Inequality t t') thus ?thesis using wf_ineq_append''[OF **] by simp
  qed
qed (auto dest: wf_subst_apply)

lemma (in intruder_model) wf_equality:
  assumes wf: "wf\<^sub>s\<^sub>t V (S@\<langle>t \<doteq> t'\<rangle>#S')"
  and \<delta>: "mgu t t' = Some \<delta>"
  and disj: "bvars\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S') \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
  shows "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
using assms
proof (induction S' arbitrary: V rule: List.rev_induct)
  case Nil thus ?case using wf_prefix[of V S "[\<langle>t \<doteq> t'\<rangle>]"] wf_subst_apply[of V S \<delta>] by auto
next
  case (snoc x S' V)
  hence "fv t' \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t S"
    using wf_eq_fv[of V, of S t t' "S'@[x]"] by auto
  hence "fv t' \<subseteq> V \<union> nonineqvars\<^sub>s\<^sub>t S"
    using vars_snd_rcv_strand_subset2(4)[of S] by blast
  hence "fv t' \<subseteq> V \<union> nonineqvars\<^sub>s\<^sub>t (S@S')" by force
  moreover have disj':
      "bvars\<^sub>s\<^sub>t (S@S') \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
      "set (bvars\<^sub>s\<^sub>t\<^sub>p x) \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
      "bvars\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S') \<inter> (subst_domain \<delta> \<union> fv_subst_range \<delta>) = {}"
    using snoc.prems(3) by auto
  ultimately have "fv (t' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
    by (metis inf_sup_aci(5) subst_apply_fv_subset_strand_trm2)
  moreover have "fv (t \<cdot> \<delta>) = fv (t' \<cdot> \<delta>)"
    by (metis MGU_is_Unifier[OF mgu_gives_MGU[OF \<delta>]])
  ultimately have *: "fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
    by simp

  from snoc.prems(1) have "wf\<^sub>s\<^sub>t V (S@\<langle>t \<doteq> t'\<rangle>#S')" using wf_prefix[of V "S@\<langle>t \<doteq> t'\<rangle>#S'"] by simp
  hence **: "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)" by (metis snoc.IH \<delta> disj'(3))

  from snoc.prems(1) have ***: "wf\<^sub>s\<^sub>t (V \<union> wfvarsoccs\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S')) [x]"
    using wf_append_exec[of V "(S@\<langle>t \<doteq> t'\<rangle>#S')" "[x]"] by simp

  show ?case
  proof (cases x)
    case (Send t)
    thus ?thesis using wf_snd_append[of "fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)" "(S@S') \<cdot>\<^sub>s\<^sub>t \<delta>"] ** by auto
  next
    case (Receive s)
    hence "fv\<^sub>s\<^sub>t\<^sub>p x \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S')" using *** by auto
    hence "fv\<^sub>s\<^sub>t\<^sub>p x \<subseteq> V \<union> nonineqvars\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S')"
      using vars_snd_rcv_strand_subset2(4)[of "S@\<langle>t \<doteq> t'\<rangle>#S'"] by blast
    hence "fv\<^sub>s\<^sub>t\<^sub>p x \<subseteq> V \<union> fv t \<union> fv t' \<union> nonineqvars\<^sub>s\<^sub>t (S@S')" by auto
    hence "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
      using subst_apply_fv_subset_strand2[of fv\<^sub>s\<^sub>t\<^sub>p]
      by (metis (no_types) inf_sup_aci(5) subst_apply_fv_union disj'(1,2))
    hence "fv\<^sub>s\<^sub>t\<^sub>p (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)" using * by blast
    hence "fv (s \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V))" using \<open>x = receive\<langle>s\<rangle>\<close> by auto
    hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) (((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)@[receive\<langle>s \<cdot> \<delta>\<rangle>])"
      using wf_rcv_append'''[OF **, of "s \<cdot> \<delta>"] by metis
    thus ?thesis using \<open>x = receive\<langle>s\<rangle>\<close> by auto
  next
    case (Equality s s')
    hence "fv s' \<subseteq> V \<union> wfvarsoccs\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S')" using *** by auto
    hence "fv s' \<subseteq> V \<union> nonineqvars\<^sub>s\<^sub>t (S@\<langle>t \<doteq> t'\<rangle>#S')"
      using vars_snd_rcv_strand_subset2(4)[of "S@\<langle>t \<doteq> t'\<rangle>#S'"] by blast
    hence "fv s' \<subseteq> V \<union> fv t \<union> fv t' \<union> nonineqvars\<^sub>s\<^sub>t (S@S')" by auto
    moreover have "fv s' = fv_r\<^sub>e\<^sub>q x" "fv (s' \<cdot> \<delta>) = fv_r\<^sub>e\<^sub>q (x \<cdot>\<^sub>s\<^sub>t\<^sub>p \<delta>)" using Equality by simp_all
    ultimately have "fv (s' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> fv (t \<cdot> \<delta>) \<union> fv (t' \<cdot> \<delta>) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)"
      using subst_apply_fv_subset_strand2[of fv_r\<^sub>e\<^sub>q x]
      by (metis (no_types) inf_sup_aci(5) subst_apply_fv_union disj'(1,2))
    hence "fv (s' \<cdot> \<delta>) \<subseteq> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V) \<union> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)" using * by blast
    hence "fv (s' \<cdot> \<delta>) \<subseteq> nonineqvars\<^sub>s\<^sub>t ((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>) \<union> fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)" using \<open>x = \<langle>s \<doteq> s'\<rangle>\<close> by auto
    hence "wf\<^sub>s\<^sub>t (fv\<^sub>s\<^sub>e\<^sub>t (\<delta> ` V)) (((S@S') \<cdot>\<^sub>s\<^sub>t \<delta>)@[\<langle>(s \<cdot> \<delta>) \<doteq> (s' \<cdot> \<delta>)\<rangle>])"
      using wf_eq_append'''[OF **] by metis
    thus ?thesis using \<open>x = \<langle>s \<doteq> s'\<rangle>\<close> by auto
  next
    case (Inequality s s') thus ?thesis using wf_ineq_append''[OF **] by simp
  qed
qed

lemma wf_rcv_prefix_ground:
  "wf\<^sub>s\<^sub>t {} ((map Receive M)@S) \<Longrightarrow> vars\<^sub>s\<^sub>t (map Receive M) = {}"
by (induct M) auto

lemma simple_wfvarsoccs\<^sub>s\<^sub>t_is_fv\<^sub>s\<^sub>n\<^sub>d:
  assumes "simple S"
  shows "wfvarsoccs\<^sub>s\<^sub>t S = \<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)"
using assms unfolding simple_def
proof (induction S)
  case (Cons x S) thus ?case by (cases x) auto
qed simp

lemma wf\<^sub>s\<^sub>t_simple_induct[consumes 2, case_names Nil ConsSnd ConsRcv ConsIneq]:
  fixes S::"('a,'b) strand"
  assumes "wf\<^sub>s\<^sub>t V S" "simple S"
          "P []"
          "\<And>v S. \<lbrakk>wf\<^sub>s\<^sub>t V S; simple S; P S\<rbrakk> \<Longrightarrow> P (S@[send\<langle>Var v\<rangle>])"
          "\<And>t S. \<lbrakk>wf\<^sub>s\<^sub>t V S; simple S; P S; fv t \<subseteq> V \<union> \<Union>set (map fv\<^sub>s\<^sub>n\<^sub>d S)\<rbrakk> \<Longrightarrow> P (S@[receive\<langle>t\<rangle>])"
          "\<And>X t t' S. \<lbrakk>wf\<^sub>s\<^sub>t V S; simple S; P S\<rbrakk> \<Longrightarrow> P (S@[\<nexists>X\<langle>t \<doteq> t'\<rangle>])"
  shows "P S"
using assms
proof (induction S rule: wf\<^sub>s\<^sub>t_induct')
  case (ConsSnd t S)
  hence "P S" by auto
  obtain v where "t = Var v" using simple_snd_is_var[OF _ \<open>simple (S@[send\<langle>t\<rangle>])\<close>] by auto
  thus ?case using ConsSnd.prems(3)[OF \<open>wf\<^sub>s\<^sub>t V S\<close> _ \<open>P S\<close>] \<open>simple (S@[send\<langle>t\<rangle>])\<close> by auto
next
  case (ConsRcv t S) thus ?case using simple_wfvarsoccs\<^sub>s\<^sub>t_is_fv\<^sub>s\<^sub>n\<^sub>d[of "S@[receive\<langle>t\<rangle>]"] by auto
qed (auto simp add: simple_def)

lemma wf_trm_stp_dom_fv_disjoint:
  "\<lbrakk>wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r S \<theta>; t \<in> trms\<^sub>s\<^sub>t S\<rbrakk> \<Longrightarrow> subst_domain \<theta> \<inter> fv t = {}"
unfolding wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r_def by force

lemma wf_constr_bvars_disj: "wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r S \<theta> \<Longrightarrow> (subst_domain \<theta> \<union> fv_subst_range \<theta>) \<inter> bvars\<^sub>s\<^sub>t S = {}"
by fastforce

lemma wf_constr_bvars_disj':
  assumes "wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r S \<theta>" "subst_domain \<delta> \<union> fv_subst_range \<delta> \<subseteq> fv\<^sub>s\<^sub>t S"
  shows "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}" (is ?A)
  and "(subst_domain \<theta> \<union> fv_subst_range \<theta>) \<inter> bvars\<^sub>s\<^sub>t (S \<cdot>\<^sub>s\<^sub>t \<delta>) = {}" (is ?B)
proof -
  have *: "(subst_domain \<theta> \<union> fv_subst_range \<theta>) \<inter> bvars\<^sub>s\<^sub>t S = {}" "fv\<^sub>s\<^sub>t S \<inter> bvars\<^sub>s\<^sub>t S = {}"
    using assms(1) unfolding wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r_def by fastforce+
  thus ?A and ?B using assms(2) bvars_subst_ident[of S \<delta>] by blast+
qed

lemma (in intruder_model) wf_trp_ik_dom_fv_disjoint: 
  assumes "wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r S \<theta>" "t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)"
  shows "subst_domain \<theta> \<inter> fv t = {}"
using \<open>wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r S \<theta>\<close> fv_subset_if_in_strand_ik'[OF \<open>t \<in> trp\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t S)\<close>]
unfolding wf\<^sub>c\<^sub>o\<^sub>n\<^sub>s\<^sub>t\<^sub>r_def by blast

lemma (in intruder_model) wf_simple_strand_first_snd_var_split:
  assumes "wf\<^sub>s\<^sub>t {} S" "simple S" "\<exists>v \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> = \<I> v"
  shows "\<exists>v S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>Var v\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> = \<I> v
                      \<and> \<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> = \<I> w)"
using assms
proof (induction S rule: wf\<^sub>s\<^sub>t_simple_induct)
  case (ConsSnd v S) show ?case
  proof (cases "\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> = \<I> w")
    case True thus ?thesis using ConsSnd.IH by fastforce
  next
    case False thus ?thesis using ConsSnd.prems by auto
  qed
next
  case (ConsRcv t' S)
  have "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S" using ConsRcv.hyps(3) vars_snd_rcv_strand_subset2(1) by force
  hence "\<exists>v \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> = \<I> v" using ConsRcv.prems(1) by fastforce
  hence "\<exists>v S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>Var v\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> = \<I> v \<and> \<not>(\<exists>w\<in>nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> = \<I> w)"
    by (metis ConsRcv.IH)
  thus ?case by fastforce 
next
  case (ConsIneq X s s' S)
  moreover have "nonineqvars\<^sub>s\<^sub>t (S @ [\<nexists>X\<langle>s \<doteq> s'\<rangle>]) = nonineqvars\<^sub>s\<^sub>t S" by auto
  ultimately have
      "\<exists>v S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>Var v\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> = \<I> v \<and> \<not>(\<exists>w\<in>nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> = \<I> w)"
    by blast
  thus ?case by fastforce
qed simp

lemma (in intruder_model) wf_strand_first_snd_var_split:
  assumes "wf\<^sub>s\<^sub>t {} S" "\<exists>v \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> \<sqsubseteq> \<I> v"
  shows "\<exists>S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f. \<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> \<sqsubseteq> \<I> w)
            \<and> ((\<exists>t'. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>t'\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>)
               \<or> (\<exists>t' t''. S = S\<^sub>p\<^sub>r\<^sub>e@\<langle>t' \<doteq> t''\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>))"
using assms
proof (induction S rule: wf\<^sub>s\<^sub>t_induct')
  case (ConsSnd t' S) show ?case
  proof (cases "\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> \<sqsubseteq> \<I> w")
    case True
    then obtain S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f where
            "\<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> \<sqsubseteq> \<I> w)"
            "(\<exists>t'. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>t'\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>) \<or>
             (\<exists>t' t''. S = S\<^sub>p\<^sub>r\<^sub>e@\<langle>t' \<doteq> t''\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>)"
      using ConsSnd.IH by moura
    thus ?thesis by fastforce
  next
    case False
    then obtain v where v: "v \<in> fv t'" "t \<cdot> \<I> \<sqsubseteq> \<I> v" using ConsSnd.prems by auto
    hence "t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>"
      using subst_mono[of "Var v" t' \<I>] vars_iff_subtermeq[of v t'] subtermeq_trans
      by auto 
    thus ?thesis using False v by auto
  qed
next
  case (ConsRcv t' S)
  have "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t S" using ConsRcv.hyps vars_snd_rcv_strand_subset2(4)[of S] by blast
  hence "\<exists>v \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> \<sqsubseteq> \<I> v" using ConsRcv.prems by fastforce
  then obtain S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f where
            "\<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> \<sqsubseteq> \<I> w)"
            "(\<exists>t'. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>t'\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>) \<or>
             (\<exists>t' t''. S = S\<^sub>p\<^sub>r\<^sub>e@\<langle>t' \<doteq> t''\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>)"
      using ConsRcv.IH by moura
  thus ?case by fastforce
next
  case (ConsEq s s' S)
  have *: "fv s' \<subseteq> nonineqvars\<^sub>s\<^sub>t S" using ConsEq.hyps vars_snd_rcv_strand_subset2(4)[of S] by blast
  show ?case
  proof (cases "\<exists>v \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> \<sqsubseteq> \<I> v")
    case True
    then obtain S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f where
            "\<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> \<sqsubseteq> \<I> w)"
            "(\<exists>t'. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>t'\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>) \<or>
             (\<exists>t' t''. S = S\<^sub>p\<^sub>r\<^sub>e@\<langle>t' \<doteq> t''\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>)"
      using ConsEq.IH by moura
    thus ?thesis by fastforce
  next
    case False
    then obtain v where "v \<in> fv s" "t \<cdot> \<I> \<sqsubseteq> \<I> v" using ConsEq.prems * by auto
    hence "t \<cdot> \<I> \<sqsubseteq> s \<cdot> \<I>"
      using vars_iff_subtermeq[of v s] subst_mono[of "Var v" s \<I>] subtermeq_trans
      by auto
    thus ?thesis using False by fastforce
  qed
next
  case (ConsIneq X s s' S)
  hence "\<exists>v \<in> nonineqvars\<^sub>s\<^sub>t S. t \<cdot> \<I> \<sqsubseteq> \<I> v" by fastforce
  then obtain S\<^sub>p\<^sub>r\<^sub>e S\<^sub>s\<^sub>u\<^sub>f where
            "\<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t S\<^sub>p\<^sub>r\<^sub>e. t \<cdot> \<I> \<sqsubseteq> \<I> w)"
            "(\<exists>t'. S = S\<^sub>p\<^sub>r\<^sub>e@send\<langle>t'\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>) \<or>
             (\<exists>t' t''. S = S\<^sub>p\<^sub>r\<^sub>e@\<langle>t' \<doteq> t''\<rangle>#S\<^sub>s\<^sub>u\<^sub>f \<and> t \<cdot> \<I> \<sqsubseteq> t' \<cdot> \<I>)"
      using ConsIneq.IH by moura
  thus ?case by fastforce
qed simp


subsection {* Constraint Semantics *}
context intruder_model
begin

subsubsection {* Definitions *}
type_synonym ('a,'b) constrsem = "('a,'b) subst \<Rightarrow> bool"

fun strand_sem_c::"('fun,'var) terms \<Rightarrow> ('fun,'var) strand \<Rightarrow> ('fun,'var) constrsem" ("\<lbrakk>_; _\<rbrakk>\<^sub>c")
where
  "\<lbrakk>M; []\<rbrakk>\<^sub>c = (\<lambda>\<I>. True)"
| "\<lbrakk>M; send\<langle>t\<rangle>#S\<rbrakk>\<^sub>c = (\<lambda>\<I>. M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c t \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>c \<I>)"
| "\<lbrakk>M; receive\<langle>t\<rangle>#S\<rbrakk>\<^sub>c = \<lbrakk>insert t M; S\<rbrakk>\<^sub>c"
| "\<lbrakk>M; \<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>c = (\<lambda>\<I>. t \<cdot> \<I> = t' \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>c \<I>)"
| "\<lbrakk>M; \<nexists>X\<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>c = (\<lambda>\<I>. ineq_model \<I> X t t' \<and> \<lbrakk>M; S\<rbrakk>\<^sub>c \<I>)"

definition constr_sem_c ("_ \<Turnstile>\<^sub>c \<langle>_,_\<rangle>") where "\<I> \<Turnstile>\<^sub>c \<langle>S,\<theta>\<rangle> \<equiv> (\<theta> supports \<I> \<and> \<lbrakk>{}; S\<rbrakk>\<^sub>c \<I>)"
abbreviation constr_sem_c' ("_ \<Turnstile>\<^sub>c \<langle>_\<rangle>" 90) where "\<I> \<Turnstile>\<^sub>c \<langle>S\<rangle> \<equiv> \<I> \<Turnstile>\<^sub>c \<langle>S,Var\<rangle>"

fun strand_sem_d::"('fun,'var) terms \<Rightarrow> ('fun,'var) strand \<Rightarrow> ('fun,'var) constrsem" ("\<lbrakk>_; _\<rbrakk>\<^sub>d")
where
  "\<lbrakk>M; []\<rbrakk>\<^sub>d = (\<lambda>\<I>. True)"
| "\<lbrakk>M; send\<langle>t\<rangle>#S\<rbrakk>\<^sub>d = (\<lambda>\<I>. M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile> t \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>d \<I>)"
| "\<lbrakk>M; receive\<langle>t\<rangle>#S\<rbrakk>\<^sub>d = \<lbrakk>insert t M; S\<rbrakk>\<^sub>d"
| "\<lbrakk>M; \<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>d = (\<lambda>\<I>. t \<cdot> \<I> = t' \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>d \<I>)"
| "\<lbrakk>M; \<nexists>X\<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>d = (\<lambda>\<I>. ineq_model \<I> X t t' \<and> \<lbrakk>M; S\<rbrakk>\<^sub>d \<I>)"

definition constr_sem_d ("_ \<Turnstile> \<langle>_,_\<rangle>") where "\<I> \<Turnstile> \<langle>S,\<theta>\<rangle> \<equiv> (\<theta> supports \<I> \<and> \<lbrakk>{}; S\<rbrakk>\<^sub>d \<I>)"
abbreviation constr_sem_d' ("_ \<Turnstile> \<langle>_\<rangle>" 90) where "\<I> \<Turnstile> \<langle>S\<rangle> \<equiv> \<I> \<Turnstile> \<langle>S,Var\<rangle>"

lemmas strand_sem_induct = strand_sem_c.induct[case_names Nil ConsSnd ConsRcv ConsEq ConsIneq]

declare constr_sem_c_def[simp]
declare constr_sem_d_def[simp]

subsubsection {* Lemmas *}
lemma strand_sem_d_mono_ik: "\<lbrakk>M \<subseteq> M'; \<lbrakk>M; S\<rbrakk>\<^sub>d \<theta>\<rbrakk> \<Longrightarrow> \<lbrakk>M'; S\<rbrakk>\<^sub>d \<theta>"
proof (induction M S arbitrary: M' rule: strand_sem_induct)
  case (ConsSnd M t S)
  hence "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile> t \<cdot> \<theta>" "\<lbrakk>M'; S\<rbrakk>\<^sub>d \<theta>" by auto
  hence "M' \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile> t \<cdot> \<theta>"
    using ideduct_mono \<open>M \<subseteq> M'\<close>
    by (metis image_mono)
  thus ?case using \<open>\<lbrakk>M'; S\<rbrakk>\<^sub>d \<theta>\<close> by simp
qed auto

lemma strand_sem_d_if_c: "\<I> \<Turnstile>\<^sub>c \<langle>S,\<theta>\<rangle> \<Longrightarrow> \<I> \<Turnstile> \<langle>S,\<theta>\<rangle>"
proof -
  assume *: "\<I> \<Turnstile>\<^sub>c \<langle>S,\<theta>\<rangle>"
  { fix M have "\<lbrakk>M; S\<rbrakk>\<^sub>c \<I> \<Longrightarrow> \<lbrakk>M; S\<rbrakk>\<^sub>d \<I>"
    proof (induction S rule: strand_sem_induct)
      case (ConsSnd M t S)
      hence "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c t \<cdot> \<I>" "\<lbrakk>M; S\<rbrakk>\<^sub>d \<I>" by auto
      thus ?case
        using strand_sem_d.simps(2)[of M t S]
        by auto
    qed (auto simp add: ineq_model_def)
  }
  thus ?thesis using * by simp
qed

lemma strand_sem_mono_ik: "\<lbrakk>M \<subseteq> M'; \<lbrakk>M; S\<rbrakk>\<^sub>c \<theta>\<rbrakk> \<Longrightarrow> \<lbrakk>M'; S\<rbrakk>\<^sub>c \<theta>"
proof (induction M S arbitrary: M' rule: strand_sem_induct)
  case (ConsSnd M t S)
  hence "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile>\<^sub>c t \<cdot> \<theta>" "\<lbrakk>M'; S\<rbrakk>\<^sub>c \<theta>" by auto
  hence "M' \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile>\<^sub>c t \<cdot> \<theta>"
    using ideduct_synth_mono \<open>M \<subseteq> M'\<close>
    by (metis image_mono)
  thus ?case using \<open>\<lbrakk>M'; S\<rbrakk>\<^sub>c \<theta>\<close> by simp
qed auto

lemma strand_sem_split_left[dest]: "\<lbrakk>M; S@S'\<rbrakk>\<^sub>c \<theta> \<Longrightarrow> \<lbrakk>M; S\<rbrakk>\<^sub>c \<theta>"
proof (induct S arbitrary: M)
  case (Cons x S) thus ?case by (cases x) simp_all
qed simp

lemma strand_sem_split_left_d[dest]: "\<lbrakk>M; S@S'\<rbrakk>\<^sub>d \<theta> \<Longrightarrow> \<lbrakk>M; S\<rbrakk>\<^sub>d \<theta>"
proof (induct S arbitrary: M)
  case (Cons x S) thus ?case by (cases x) simp_all
qed simp

lemma strand_sem_split_right[dest]: "\<lbrakk>M; S@S'\<rbrakk>\<^sub>c \<theta> \<Longrightarrow> \<lbrakk>M \<union> ik\<^sub>s\<^sub>t S; S'\<rbrakk>\<^sub>c \<theta>"
proof (induction S arbitrary: M rule: ik\<^sub>s\<^sub>t_induct)
  case (ConsRcv t S M) thus ?case using ConsRcv.IH[of "insert t M"] by simp
qed simp_all

lemma strand_sem_split_right_d[dest]: "\<lbrakk>M; S@S'\<rbrakk>\<^sub>d \<theta> \<Longrightarrow> \<lbrakk>M \<union> ik\<^sub>s\<^sub>t S; S'\<rbrakk>\<^sub>d \<theta>"
proof (induction S arbitrary: M rule: ik\<^sub>s\<^sub>t_induct)
  case (ConsRcv t S M) thus ?case using ConsRcv.IH[of "insert t M"] by simp
qed simp_all

lemma strand_sem_snd_split[dest]:
  "\<lbrakk>\<lbrakk>M; map Send X\<rbrakk>\<^sub>c \<theta>; x \<in> set X\<rbrakk> \<Longrightarrow> \<lbrakk>M; [send\<langle>x\<rangle>]\<rbrakk>\<^sub>c \<theta>"
by (induct "map Send X" arbitrary: X rule: strand_sem_c.induct) auto

lemma strand_sem_snd_split_d[dest]:
  "\<lbrakk>\<lbrakk>M; map Send X\<rbrakk>\<^sub>d \<theta>; x \<in> set X\<rbrakk> \<Longrightarrow> \<lbrakk>M; [send\<langle>x\<rangle>]\<rbrakk>\<^sub>d \<theta>"
by (induct "map Send X" arbitrary: X rule: strand_sem_d.induct) auto

lemma strand_sem_snd_split'[dest]:
  "\<lbrakk>\<lbrakk>M; map Send X@S\<rbrakk>\<^sub>c \<theta>; x \<in> set X\<rbrakk> \<Longrightarrow> \<lbrakk>M; send\<langle>x\<rangle>#S\<rbrakk>\<^sub>c \<theta>"
using strand_sem_snd_split strand_sem_split_right strand_sem_split_left
by (metis ik_snd_empty strand_sem_c.simps(2) sup_bot.right_neutral)

lemma strand_sem_append[intro]: "\<lbrakk>\<lbrakk>M; S\<rbrakk>\<^sub>c \<theta>; \<lbrakk>M \<union> ik\<^sub>s\<^sub>t S; S'\<rbrakk>\<^sub>c \<theta>\<rbrakk> \<Longrightarrow> \<lbrakk>M; S@S'\<rbrakk>\<^sub>c \<theta>"
proof (induction S arbitrary: M)
  case (Cons x S) thus ?case by (cases x) auto
qed simp

lemma strand_sem_append_d[intro]: "\<lbrakk>\<lbrakk>M; S\<rbrakk>\<^sub>d \<theta>; \<lbrakk>M \<union> ik\<^sub>s\<^sub>t S; S'\<rbrakk>\<^sub>d \<theta>\<rbrakk> \<Longrightarrow> \<lbrakk>M; S@S'\<rbrakk>\<^sub>d \<theta>"
proof (induction S arbitrary: M)
  case (Cons x S) thus ?case by (cases x) auto
qed simp

lemma strand_sem_subst:
  assumes "(subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "\<lbrakk>M; S\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>) \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>"
using assms
proof (induction S arbitrary: \<delta> M rule: strand_sem_induct)
  case (ConsSnd M t S)
  hence "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>" "M \<cdot>\<^sub>s\<^sub>e\<^sub>t (\<delta> \<circ>\<^sub>s \<theta>) \<turnstile>\<^sub>c t \<cdot> (\<delta> \<circ>\<^sub>s \<theta>)" by auto
  hence "(M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile>\<^sub>c (t \<cdot> \<delta>) \<cdot> \<theta>"
    using subst_comp_all[of \<delta> \<theta> M] subst_comp[of t \<delta> \<theta>] by simp
  thus ?case using \<open>\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>\<close> unfolding subst_apply_strand_def by simp
next
  case (ConsRcv M t S)
  have *: "\<lbrakk>insert t M; S\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)" using ConsRcv.prems(1) by simp
  have "bvars\<^sub>s\<^sub>t (receive\<langle>t\<rangle>#S) = bvars\<^sub>s\<^sub>t S" by auto
  hence **: "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}" using ConsRcv.prems(2) by blast
  have "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; receive\<langle>t \<cdot> \<delta>\<rangle>#(S \<cdot>\<^sub>s\<^sub>t \<delta>)\<rbrakk>\<^sub>c \<theta>"
    using ConsRcv.IH[OF * **] by (simp add: subst_all_insert)
  thus ?case by simp
next
  case (ConsIneq M X t t' S)
  hence *: "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>" and
        ***: "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> set X = {}" 
    unfolding bvars\<^sub>s\<^sub>t_def ineq_model_def by auto
  have **: "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> t \<cdot> \<gamma> \<cdot> \<delta> \<cdot> \<theta> \<noteq> t' \<cdot> \<gamma> \<cdot> \<delta> \<cdot> \<theta>"
    using ConsIneq by (simp add: ineq_model_def)
  have "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>)
          \<longrightarrow> (subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> (subst_domain \<gamma> \<union> fv_subst_range \<gamma>) = {}"
    using * ** *** by (auto simp del: subst_domain_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> \<gamma> \<circ>\<^sub>s \<delta> = \<delta> \<circ>\<^sub>s \<gamma>"
    by (metis subst_comp_eq_if_disjoint_vars)
  hence "ineq_model \<theta> X (t \<cdot> \<delta>) (t' \<cdot> \<delta>)"
    using ** unfolding ineq_model_def by (metis (no_types, hide_lams) subst_comp)
  moreover have "rm_vars (set X) \<delta> = \<delta>" using ConsIneq.prems(2) by force
  ultimately show ?case using * by auto
qed simp_all

lemma strand_sem_subst':
  assumes "(subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta> \<Longrightarrow> \<lbrakk>M; S\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)"
using assms
proof (induction S arbitrary: \<delta> M rule: strand_sem_induct)
  case (ConsSnd M t S)
  hence "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; [send\<langle>t\<rangle>] \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>" "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>" by auto
  hence "\<lbrakk>M; S\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)" using ConsSnd.IH[OF _] ConsSnd.prems(2) by auto
  moreover have "\<lbrakk>M; [send\<langle>t\<rangle>]\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)"
  proof -
    have "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta> \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile>\<^sub>c t \<cdot> \<delta> \<cdot> \<theta>" using \<open>\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; [send\<langle>t\<rangle>] \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>\<close> by auto
    hence "M \<cdot>\<^sub>s\<^sub>e\<^sub>t (\<delta> \<circ>\<^sub>s \<theta>) \<turnstile>\<^sub>c t \<cdot> (\<delta> \<circ>\<^sub>s \<theta>)" using subst_comp subst_comp_all by metis
    thus "\<lbrakk>M; [send\<langle>t\<rangle>]\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)" by auto
  qed
  ultimately show ?case by auto
next
  case (ConsRcv M t S)
  hence "\<lbrakk>(insert t M) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c \<theta>" by (simp add: subst_all_insert)
  thus ?case using ConsRcv.IH ConsRcv.prems(2) by auto
next
  case (ConsIneq M X t t' S)
  have "rm_vars (set X) \<delta> = \<delta>" using ConsIneq.prems(2) by force
  hence *: "\<lbrakk>M; S\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)" and
        **: "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> t \<cdot> \<delta> \<cdot> \<gamma> \<cdot> \<theta> \<noteq> t' \<cdot> \<delta> \<cdot> \<gamma> \<cdot> \<theta>" and
        ***: "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> set X = {}"
    using ConsIneq unfolding bvars\<^sub>s\<^sub>t_def by (auto simp add: ineq_model_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>)
          \<longrightarrow> (subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> (subst_domain \<gamma> \<union> fv_subst_range \<gamma>) = {}"
    by (auto simp del: subst_domain_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> \<gamma> \<circ>\<^sub>s \<delta> = \<delta> \<circ>\<^sub>s \<gamma>"
    by (metis subst_comp_eq_if_disjoint_vars)
  hence "ineq_model (\<delta> \<circ>\<^sub>s \<theta>) X t t'"
    using ** unfolding ineq_model_def by (metis (no_types, hide_lams) subst_comp)
  thus ?case using * by auto 
next
  case ConsEq thus ?case unfolding bvars\<^sub>s\<^sub>t_def by auto
qed simp_all

lemma strand_sem_subst_d:
  assumes "(subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "\<lbrakk>M; S\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>) \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>"
using assms
proof (induction S arbitrary: \<delta> M rule: strand_sem_induct)
  case (ConsSnd M t S)
  hence "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>" "M \<cdot>\<^sub>s\<^sub>e\<^sub>t (\<delta> \<circ>\<^sub>s \<theta>) \<turnstile> t \<cdot> (\<delta> \<circ>\<^sub>s \<theta>)" by auto
  hence "(M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile> (t \<cdot> \<delta>) \<cdot> \<theta>"
    using subst_comp_all[of \<delta> \<theta> M] subst_comp[of t \<delta> \<theta>] by simp
  thus ?case using \<open>\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>\<close> by simp
next
  case (ConsRcv M t S) 
  have *: "\<lbrakk>insert t M; S\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>)" using ConsRcv.prems(1) by simp
  have "bvars\<^sub>s\<^sub>t (receive\<langle>t\<rangle>#S) = bvars\<^sub>s\<^sub>t S" by auto
  hence **: "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> bvars\<^sub>s\<^sub>t S = {}" using ConsRcv.prems(2) by blast
  have "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; receive\<langle>t \<cdot> \<delta>\<rangle>#(S \<cdot>\<^sub>s\<^sub>t \<delta>)\<rbrakk>\<^sub>d \<theta>"
    using ConsRcv.IH[OF * **] by (simp add: subst_all_insert)
  thus ?case by simp
next
  case (ConsIneq M X t t' S)
  hence *: "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>" and
        **: "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> t \<cdot> \<gamma> \<cdot> \<delta> \<cdot> \<theta> \<noteq> t' \<cdot> \<gamma> \<cdot> \<delta> \<cdot> \<theta>" and
        ***: "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> set X = {}"
    unfolding bvars\<^sub>s\<^sub>t_def by (auto simp add: ineq_model_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>)
          \<longrightarrow> (subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> (subst_domain \<gamma> \<union> fv_subst_range \<gamma>) = {}"
    by (auto simp del: subst_domain_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> \<gamma> \<circ>\<^sub>s \<delta> = \<delta> \<circ>\<^sub>s \<gamma>"
    by (metis subst_comp_eq_if_disjoint_vars)
  hence "ineq_model \<theta> X (t \<cdot> \<delta>) (t' \<cdot> \<delta>)"
    using ** unfolding ineq_model_def by (metis (no_types, hide_lams) subst_comp)
  moreover have "rm_vars (set X) \<delta> = \<delta>" using ConsIneq.prems(2) by force
  ultimately show ?case using * by auto
next
  case ConsEq thus ?case unfolding bvars\<^sub>s\<^sub>t_def by auto
qed simp_all

lemma strand_sem_subst_d':
  assumes "(subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta> \<Longrightarrow> \<lbrakk>M; S\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>)"
using assms
proof (induction S arbitrary: \<delta> M rule: strand_sem_induct)
  case (ConsSnd M t S)
  hence "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; [send\<langle>t\<rangle>] \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>" "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>" by auto
  hence "\<lbrakk>M; S\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>)" using ConsSnd.IH[OF _] ConsSnd.prems(2) by auto
  moreover have "\<lbrakk>M; [send\<langle>t\<rangle>]\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>)"
  proof -
    have "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta> \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> \<turnstile> t \<cdot> \<delta> \<cdot> \<theta>" using \<open>\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; [send\<langle>t\<rangle>] \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>\<close> by auto
    hence "M \<cdot>\<^sub>s\<^sub>e\<^sub>t (\<delta> \<circ>\<^sub>s \<theta>) \<turnstile> t \<cdot> (\<delta> \<circ>\<^sub>s \<theta>)" using subst_comp subst_comp_all by metis
    thus "\<lbrakk>M; [send\<langle>t\<rangle>]\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>)" by auto
  qed
  ultimately show ?case by auto
next
  case (ConsRcv M t S)
  hence "\<lbrakk>(insert t M) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>d \<theta>" by (simp add: subst_all_insert)
  thus ?case using ConsRcv.IH ConsRcv.prems(2) by auto
next
  case (ConsIneq M X t t' S)
  have "rm_vars (set X) \<delta> = \<delta>" using ConsIneq.prems(2) by force
  hence *: "\<lbrakk>M; S\<rbrakk>\<^sub>d (\<delta> \<circ>\<^sub>s \<theta>)" and
        **: "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> t \<cdot> \<delta> \<cdot> \<gamma> \<cdot> \<theta> \<noteq> t' \<cdot> \<delta> \<cdot> \<gamma> \<cdot> \<theta>" and
        ***: "(subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> set X = {}"
    using ConsIneq by (auto simp add: ineq_model_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>)
          \<longrightarrow> (subst_domain \<delta> \<union> fv_subst_range \<delta>) \<inter> (subst_domain \<gamma> \<union> fv_subst_range \<gamma>) = {}"
    by (auto simp del: subst_domain_def)
  hence "\<forall>\<gamma>. subst_domain \<gamma> = set X \<and> ground (subst_range \<gamma>) \<longrightarrow> \<gamma> \<circ>\<^sub>s \<delta> = \<delta> \<circ>\<^sub>s \<gamma>"
    by (metis subst_comp_eq_if_disjoint_vars)
  hence "ineq_model (\<delta> \<circ>\<^sub>s \<theta>) X t t'"
    using ** unfolding ineq_model_def by (metis (no_types, hide_lams) subst_comp)
  thus ?case using * by auto
next
  case ConsEq thus ?case unfolding bvars\<^sub>s\<^sub>t_def by auto
qed simp_all

lemma strand_sem_subst_Idem:
  assumes "(subst_domain \<delta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<delta>)) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  shows "\<lbrakk>\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>; S \<cdot>\<^sub>s\<^sub>t \<delta>\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>); Idem \<delta>\<rbrakk> \<Longrightarrow> \<lbrakk>M; S\<rbrakk>\<^sub>c (\<delta> \<circ>\<^sub>s \<theta>)"
apply (induct "S \<cdot>\<^sub>s\<^sub>t \<delta>" rule: strand_sem_c.induct)
by (metis Idem_def comp_assoc strand_sem_subst'[OF assms])+

lemma strand_sem_subst_comp:
  assumes "(subst_domain \<theta> \<union> fv\<^sub>s\<^sub>e\<^sub>t (subst_range \<theta>)) \<inter> bvars\<^sub>s\<^sub>t S = {}"
  and "\<lbrakk>M; S\<rbrakk>\<^sub>c \<delta>" "subst_domain \<theta> \<inter> (vars\<^sub>s\<^sub>t S \<union> fv\<^sub>s\<^sub>e\<^sub>t M) = {}"
  shows "\<lbrakk>M; S\<rbrakk>\<^sub>c (\<theta> \<circ>\<^sub>s \<delta>)"
proof -
  from assms(3) have "subst_domain \<theta> \<inter> vars\<^sub>s\<^sub>t S = {}" "subst_domain \<theta> \<inter> fv\<^sub>s\<^sub>e\<^sub>t M = {}" by auto
  hence "S \<cdot>\<^sub>s\<^sub>t \<theta> = S" "M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta> = M" using strand_substI set_subst_ident[of M \<theta>] by (blast, blast)
  thus ?thesis using assms(2) by (auto simp add: strand_sem_subst'[OF assms(1)])
qed

lemma strand_sem_d_imp_ineqs_neq:
  assumes "\<lbrakk>M; S\<rbrakk>\<^sub>d \<I>" "\<nexists>X\<langle>t \<doteq> t'\<rangle> \<in> set S"
  shows "t \<noteq> t'"
using assms
proof (induction rule: strand_sem_induct)
  case (ConsIneq M Y s s' S) thus ?case
  proof (cases "\<nexists>X\<langle>t \<doteq> t'\<rangle> \<in> set S")
    case False
    hence "X = Y" "t = s" "t' = s'" using ConsIneq by auto
    hence "\<forall>\<theta>. subst_domain \<theta> = set X \<and> ground (subst_range \<theta>) \<longrightarrow> t \<cdot> \<theta> \<cdot> \<I> \<noteq> t' \<cdot> \<theta> \<cdot> \<I>"
      using ConsIneq by (auto simp add: ineq_model_def)
    then obtain \<theta> where "subst_domain \<theta> = set X" "ground (subst_range \<theta>)" "t \<cdot> \<theta> \<cdot> \<I> \<noteq> t' \<cdot> \<theta> \<cdot> \<I>"
      using interpretation_subst_exists'[of "set X"] by moura
    thus ?thesis by auto
  qed simp
qed simp_all

lemma strand_sem_c_imp_ineqs_neq:
  assumes "\<lbrakk>M; S\<rbrakk>\<^sub>c \<I>" "\<nexists>X\<langle>t \<doteq> t'\<rangle> \<in> set S"
  shows "t \<noteq> t' \<and> (\<forall>\<delta>. subst_domain \<delta> = set X \<and> ground (subst_range \<delta>) \<longrightarrow> t \<cdot> \<delta> \<noteq> t' \<cdot> \<delta> \<and> t \<cdot> \<delta> \<cdot> \<I> \<noteq> t' \<cdot> \<delta> \<cdot> \<I>)"
using assms
proof (induction rule: strand_sem_induct)
  case (ConsIneq M Y s s' S) thus ?case
  proof (cases "\<nexists>X\<langle>t \<doteq> t'\<rangle> \<in> set S")
    case False
    hence "X = Y" "t = s" "t' = s'" using ConsIneq by auto
    hence *: "\<forall>\<theta>. subst_domain \<theta> = set X \<and> ground (subst_range \<theta>) \<longrightarrow> t \<cdot> \<theta> \<cdot> \<I> \<noteq> t' \<cdot> \<theta> \<cdot> \<I>"
      using ConsIneq by (auto simp add: ineq_model_def)
    then obtain \<theta> where \<theta>: "subst_domain \<theta> = set X" "ground (subst_range \<theta>)" "t \<cdot> \<theta> \<cdot> \<I> \<noteq> t' \<cdot> \<theta> \<cdot> \<I>"
      using interpretation_subst_exists'[of "set X"] by moura
    hence "t \<noteq> t'" by auto
    moreover have "\<And>\<I> \<theta>. t \<cdot> \<theta> \<cdot> \<I> \<noteq> t' \<cdot> \<theta> \<cdot> \<I> \<Longrightarrow> t \<cdot> \<theta> \<noteq> t' \<cdot> \<theta>" by auto
    ultimately show ?thesis using * by auto
  qed simp
qed simp_all

lemma strand_sem_snd_map:
  assumes "\<And>t. t \<in> set X \<Longrightarrow> \<lbrakk>M; [send\<langle>t\<rangle>]\<rbrakk>\<^sub>c \<I>"
  shows "\<lbrakk>M; map Send X\<rbrakk>\<^sub>c \<I>"
using assms by (induct X) auto

lemma strand_sem_snd_map_d:
  assumes "\<And>t. t \<in> set X \<Longrightarrow> \<lbrakk>M; [send\<langle>t\<rangle>]\<rbrakk>\<^sub>d \<I>"
  shows "\<lbrakk>M; map Send X\<rbrakk>\<^sub>d \<I>"
using assms by (induct X) auto

lemma strand_sem_rcv_map: "\<lbrakk>M; map Receive X\<rbrakk>\<^sub>c \<I>" by (induct X arbitrary: M) auto

lemma strand_sem_rcv_map_d: "\<lbrakk>M; map Receive X\<rbrakk>\<^sub>d \<I>" by (induct X arbitrary: M) auto

lemma strand_sem_wf_simple_fv_sat:
  assumes "wf\<^sub>s\<^sub>t {} S" "simple S" "\<lbrakk>{}; S\<rbrakk>\<^sub>c \<I>"
  shows "\<And>v. v \<in> nonineqvars\<^sub>s\<^sub>t S \<Longrightarrow> ik\<^sub>s\<^sub>t S \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c \<I> v"
using assms
proof (induction S rule: wf\<^sub>s\<^sub>t_simple_induct)
  case (ConsRcv t S)
  have "v \<in> nonineqvars\<^sub>s\<^sub>t S"
    using ConsRcv.hyps(3) ConsRcv.prems(1) vars_snd_rcv_strand2 by fastforce
  moreover have "\<lbrakk>{}; S\<rbrakk>\<^sub>c \<I>" using \<open>\<lbrakk>{}; S@[receive\<langle>t\<rangle>]\<rbrakk>\<^sub>c \<I>\<close> by blast
  moreover have "ik\<^sub>s\<^sub>t S \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<subseteq> ik\<^sub>s\<^sub>t (S@[receive\<langle>t\<rangle>]) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>" by auto
  ultimately show ?case using ConsRcv.IH ideduct_synth_mono by meson
next
  case (ConsIneq X t t' S)
  hence "v \<in> nonineqvars\<^sub>s\<^sub>t S" by fastforce
  moreover have "\<lbrakk>{}; S\<rbrakk>\<^sub>c \<I>" using \<open>\<lbrakk>{}; S@[\<nexists>X\<langle>t \<doteq> t'\<rangle>]\<rbrakk>\<^sub>c \<I>\<close> by blast
  moreover have "ik\<^sub>s\<^sub>t S \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<subseteq> ik\<^sub>s\<^sub>t (S@[\<nexists>X\<langle>t \<doteq> t'\<rangle>]) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>" by auto
  ultimately show ?case using ConsIneq.IH ideduct_synth_mono by meson
next
  case (ConsSnd w S)
  hence *: "\<lbrakk>{}; S\<rbrakk>\<^sub>c \<I>" "ik\<^sub>s\<^sub>t S \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c \<I> w" by auto
  have **: "ik\<^sub>s\<^sub>t S \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<subseteq> ik\<^sub>s\<^sub>t (S@[send\<langle>Var w\<rangle>]) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>" by simp
  show ?case
  proof (cases "v = w")
    case True thus ?thesis using *(2) ideduct_synth_mono[OF _ **] by meson
  next
    case False
    hence "v \<in> nonineqvars\<^sub>s\<^sub>t S" using ConsSnd.prems(1) by auto
    thus ?thesis using ConsSnd.IH[OF _ *(1)] ideduct_synth_mono[OF _ **] by metis
  qed
qed simp

lemma strand_sem_wf_ik_or_eqs_rhs_fun_subterm:
  assumes "wf\<^sub>s\<^sub>t {} A" "\<lbrakk>{}; A\<rbrakk>\<^sub>c \<I>" "Var x \<in> ik\<^sub>s\<^sub>t A" "\<I> x = Fun f T"
          "t\<^sub>i \<in> set T" "\<not>ik\<^sub>s\<^sub>t A \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c t\<^sub>i" "interpretation\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<I>"
  obtains S where
    "Fun f S \<in> subterms\<^sub>s\<^sub>e\<^sub>t (ik\<^sub>s\<^sub>t A) \<or> Fun f S \<in> subterms\<^sub>s\<^sub>e\<^sub>t (eqs_rhs\<^sub>s\<^sub>t A)"
    "Fun f T = Fun f S \<cdot> \<I>"
proof -
  have "x \<in> nonineqvars\<^sub>s\<^sub>t A" using assms(3) vars_iff_subterm_or_eq by force
  moreover have "Fun f T \<cdot> \<I> = Fun f T"
    by (metis subst_ground_ident interpretation_grounds_all assms(4,7))
  ultimately obtain A\<^sub>p\<^sub>r\<^sub>e A\<^sub>s\<^sub>u\<^sub>f where *:
      "\<not>(\<exists>w \<in> nonineqvars\<^sub>s\<^sub>t A\<^sub>p\<^sub>r\<^sub>e. Fun f T \<sqsubseteq> \<I> w)"
      "(\<exists>t. A = A\<^sub>p\<^sub>r\<^sub>e@send\<langle>t\<rangle>#A\<^sub>s\<^sub>u\<^sub>f \<and> Fun f T \<sqsubseteq> t \<cdot> \<I>) \<or>
        (\<exists>t t'. A = A\<^sub>p\<^sub>r\<^sub>e@\<langle>t \<doteq> t'\<rangle>#A\<^sub>s\<^sub>u\<^sub>f \<and> Fun f T \<sqsubseteq> t \<cdot> \<I>)"
    using wf_strand_first_snd_var_split[OF assms(1)] assms(4) subtermeqI' by metis
  moreover
  { fix t assume **: "A = A\<^sub>p\<^sub>r\<^sub>e@send\<langle>t\<rangle>#A\<^sub>s\<^sub>u\<^sub>f" "Fun f T \<sqsubseteq> t \<cdot> \<I>"
    hence "ik\<^sub>s\<^sub>t A\<^sub>p\<^sub>r\<^sub>e \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c t \<cdot> \<I>" "\<not>ik\<^sub>s\<^sub>t A\<^sub>p\<^sub>r\<^sub>e \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I> \<turnstile>\<^sub>c t\<^sub>i"
      using assms(2,6) by (auto intro: ideduct_synth_mono)
    then obtain s where s: "s \<in> ik\<^sub>s\<^sub>t A\<^sub>p\<^sub>r\<^sub>e" "Fun f T \<sqsubseteq> s \<cdot> \<I>"
      using assms(5) **(2) by (induct rule: intruder_synth_induct) auto
    then obtain g S where gS: "Fun g S \<sqsubseteq> s" "Fun f T = Fun g S \<cdot> \<I>"
      using subterm_subst_not_img_subterm[OF s(2)] *(1) by force
    hence ?thesis using that **(1) s(1) by force
  }
  moreover
  { fix t t' assume **: "A = A\<^sub>p\<^sub>r\<^sub>e@\<langle>t \<doteq> t'\<rangle>#A\<^sub>s\<^sub>u\<^sub>f" "Fun f T \<sqsubseteq> t \<cdot> \<I>"
    with assms(2) have "t \<cdot> \<I> = t' \<cdot> \<I>" by auto
    hence "Fun f T \<sqsubseteq> t' \<cdot> \<I>" using **(2) by auto
    from assms(1) **(1) have "fv t' \<subseteq> nonineqvars\<^sub>s\<^sub>t A\<^sub>p\<^sub>r\<^sub>e"
      using wf_eq_fv[of "{}" A\<^sub>p\<^sub>r\<^sub>e t t' A\<^sub>s\<^sub>u\<^sub>f] vars_snd_rcv_strand_subset2(4)[of A\<^sub>p\<^sub>r\<^sub>e]
      by blast
    then obtain g S where gS: "Fun g S \<sqsubseteq> t'" "Fun f T = Fun g S \<cdot> \<I>"
      using subterm_subst_not_img_subterm[OF \<open>Fun f T \<sqsubseteq> t' \<cdot> \<I>\<close>] *(1) by fastforce
    hence ?thesis using that **(1) by auto
  }
  ultimately show ?thesis by auto
qed

lemma strand_sem_not_unif_is_sat_ineq:
  assumes "\<not>(\<exists>\<theta>. Unifier \<theta> t t')"
  shows "\<lbrakk>M; [\<nexists>X\<langle>t \<doteq> t'\<rangle>]\<rbrakk>\<^sub>c \<I>" "\<lbrakk>M; [\<nexists>X\<langle>t \<doteq> t'\<rangle>]\<rbrakk>\<^sub>d \<I>"
using assms subst_comp ineq_model_def
by (metis (mono_tags, lifting) strand_sem_c.simps(1,5),
    metis (mono_tags, lifting) strand_sem_d.simps(1,5)) 

subsection {* Constraint Semantics (slightly modified and equivalent version) *}
fun strand_sem_c'::"('fun,'var) terms \<Rightarrow> ('fun,'var) strand \<Rightarrow> ('fun,'var) constrsem" ("\<lbrakk>_; _\<rbrakk>\<^sub>c'")
where
  "\<lbrakk>M; []\<rbrakk>\<^sub>c' = (\<lambda>\<I>. True)"
| "\<lbrakk>M; send\<langle>t\<rangle>#S\<rbrakk>\<^sub>c' = (\<lambda>\<I>. M \<turnstile>\<^sub>c t \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>c' \<I>)"
| "\<lbrakk>M; receive\<langle>t\<rangle>#S\<rbrakk>\<^sub>c' = (\<lambda>\<I>. \<lbrakk>insert (t \<cdot> \<I>) M; S\<rbrakk>\<^sub>c' \<I>)"
| "\<lbrakk>M; \<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>c' = (\<lambda>\<I>. t \<cdot> \<I> = t' \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>c' \<I>)"
| "\<lbrakk>M; \<nexists>X\<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>c' = (\<lambda>\<I>. ineq_model \<I> X t t' \<and> \<lbrakk>M; S\<rbrakk>\<^sub>c' \<I>)"

fun strand_sem_d'::"('fun,'var) terms \<Rightarrow> ('fun,'var) strand \<Rightarrow> ('fun,'var) constrsem" ("\<lbrakk>_; _\<rbrakk>\<^sub>d'")
where
  "\<lbrakk>M; []\<rbrakk>\<^sub>d' = (\<lambda>\<I>. True)"
| "\<lbrakk>M; send\<langle>t\<rangle>#S\<rbrakk>\<^sub>d' = (\<lambda>\<I>. M \<turnstile> t \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>d' \<I>)"
| "\<lbrakk>M; receive\<langle>t\<rangle>#S\<rbrakk>\<^sub>d' = (\<lambda>\<I>. \<lbrakk>insert (t \<cdot> \<I>) M; S\<rbrakk>\<^sub>d' \<I>)"
| "\<lbrakk>M; \<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>d' = (\<lambda>\<I>. t \<cdot> \<I> = t' \<cdot> \<I> \<and> \<lbrakk>M; S\<rbrakk>\<^sub>d' \<I>)"
| "\<lbrakk>M; \<nexists>X\<langle>t \<doteq> t'\<rangle>#S\<rbrakk>\<^sub>d' = (\<lambda>\<I>. ineq_model \<I> X t t' \<and> \<lbrakk>M; S\<rbrakk>\<^sub>d' \<I>)"

lemma strand_sem_eq_defs:
  "\<lbrakk>M; \<A>\<rbrakk>\<^sub>c \<I> = \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>c' \<I>"
  "\<lbrakk>M; \<A>\<rbrakk>\<^sub>d \<I> = \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>d' \<I>"
proof -
  have 1: "\<lbrakk>M; \<A>\<rbrakk>\<^sub>c \<I> \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>c' \<I>"
    by (induct \<A> arbitrary: M rule: strand_sem_induct) force+
  have 2: "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>c' \<I> \<Longrightarrow> \<lbrakk>M; \<A>\<rbrakk>\<^sub>c \<I>"
    by (induct \<A> arbitrary: M rule: strand_sem_c'.induct) auto
  have 3: "\<lbrakk>M; \<A>\<rbrakk>\<^sub>d \<I> \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>d' \<I>"
    by (induct \<A> arbitrary: M rule: strand_sem_induct) force+
  have 4: "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>d' \<I> \<Longrightarrow> \<lbrakk>M; \<A>\<rbrakk>\<^sub>d \<I>"
    by (induct \<A> arbitrary: M rule: strand_sem_d'.induct) auto

  show "\<lbrakk>M; \<A>\<rbrakk>\<^sub>c \<I> = \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>c' \<I>" "\<lbrakk>M; \<A>\<rbrakk>\<^sub>d \<I> = \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<I>; \<A>\<rbrakk>\<^sub>d' \<I>"
    by (metis 1 2, metis 3 4)
qed

lemma strand_sem_split'[dest]:
  "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S@S'\<rbrakk>\<^sub>c' \<theta> \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S\<rbrakk>\<^sub>c' \<theta>"
  "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S@S'\<rbrakk>\<^sub>c' \<theta> \<Longrightarrow> \<lbrakk>(M \<union> ik\<^sub>s\<^sub>t S) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S'\<rbrakk>\<^sub>c' \<theta>"
  "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S@S'\<rbrakk>\<^sub>d' \<theta> \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S\<rbrakk>\<^sub>d' \<theta>"
  "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S@S'\<rbrakk>\<^sub>d' \<theta> \<Longrightarrow> \<lbrakk>(M \<union> ik\<^sub>s\<^sub>t S) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S'\<rbrakk>\<^sub>d' \<theta>"
using strand_sem_eq_defs[of M "S@S'" \<theta>]
      strand_sem_eq_defs[of M S \<theta>]
      strand_sem_eq_defs[of "M \<union> ik\<^sub>s\<^sub>t S" S' \<theta>]
by auto

lemma strand_sem_append'[intro]:
  "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S\<rbrakk>\<^sub>c' \<theta> \<Longrightarrow> \<lbrakk>(M \<union> ik\<^sub>s\<^sub>t S) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S'\<rbrakk>\<^sub>c' \<theta> \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S@S'\<rbrakk>\<^sub>c' \<theta>"
  "\<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S\<rbrakk>\<^sub>d' \<theta> \<Longrightarrow> \<lbrakk>(M \<union> ik\<^sub>s\<^sub>t S) \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S'\<rbrakk>\<^sub>d' \<theta> \<Longrightarrow> \<lbrakk>M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<theta>; S@S'\<rbrakk>\<^sub>d' \<theta>"
using strand_sem_eq_defs[of M "S@S'" \<theta>]
      strand_sem_eq_defs[of M S \<theta>]
      strand_sem_eq_defs[of "M \<union> ik\<^sub>s\<^sub>t S" S' \<theta>]
by auto

end

end
